#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Action Module server

This uses the `Circuits <http://circuitsframework.com/>` framework to
listen on multiple message destinations, send their events to the relevant
handlers, and acknowledge the messages when they have been processed.

"""

from __future__ import print_function

import logging
from logging.handlers import RotatingFileHandler
from six import string_types
import re
import os
import filelock
from circuits import Manager, BaseComponent, Component, Debugger, Loader
from resilient import get_config_file
from resilient import constants as resilient_constants
from resilient_circuits import constants, helpers
from resilient_circuits.app_argument_parser import AppArgumentParser
from resilient_circuits.component_loader import ComponentLoader
from resilient_circuits.actions_component import Actions, ResilientComponent


application = None
logging_initialized = False


class RedactingFilter(logging.Filter):
    """ Redacting logging filter to prevent Resilient circuits sensitive password values from being logged.

    """
    def __init__(self):
        super(RedactingFilter, self).__init__()

    def filter(self, record):
        # Best effort regex filter pattern to redact password logging.
        if isinstance(record.msg, string_types):
            for p in constants.PASSWD_PATTERNS:
                if p in record.msg.lower():
                    regex = r"{0}(?=.*?'|\":\s)(None,|.+?,|.+?'|.+\"|}})".format(p)
                    record.msg = re.sub(regex, r"***", record.msg)

        return True


# Main component for our application
class App(Component):
    """Our main app component, which sets up the Resilient services and other components"""

    FILE_LOG_FORMAT = '%(asctime)s %(levelname)s [%(module)s] [%(threadName)s] %(message)s'
    SYSLOG_LOG_FORMAT = '%(module)s: %(levelname)s %(message)s'
    STDERR_LOG_FORMAT = '%(asctime)s %(levelname)s [%(module)s] [%(threadName)s] %(message)s'

    def __init__(self, auto_load_components=True, config_file=None, ALLOW_UNRECOGNIZED=False, IS_SELFTEST=False):
        super(App, self).__init__()
        # Read the configuration options
        self.ALLOW_UNRECOGNIZED = ALLOW_UNRECOGNIZED
        resilient_constants.ALLOW_UNRECOGNIZED = ALLOW_UNRECOGNIZED
        self.IS_SELFTEST = IS_SELFTEST
        self.action_component = None
        self.component_loader = None
        self.auto_load_components = auto_load_components
        self.config_file = config_file or get_config_file()
        self.do_initialization()

    def do_initialization(self):
        self.opts = helpers.get_configs(path_config_file=self.config_file, ALLOW_UNRECOGNIZED=self.ALLOW_UNRECOGNIZED)

        self.config_logging(self.opts["logdir"], self.opts["loglevel"], self.opts["logfile"], self.opts[constants.APP_CONFIG_LOG_MAX_BYTES], self.opts[constants.APP_CONFIG_LOG_BACKUP_COUNT])
        LOG.info("Configuration file: %s", self.config_file)
        LOG.info("Resilient server: %s", self.opts.get("host"))
        if self.opts.get("email", None):
            LOG.info("Resilient user: %s", self.opts.get("email"))
        if self.opts.get("api_key_id", None):
            LOG.info("Resilient api key id: %s", self.opts.get("api_key_id"))
        if self.opts.get("api_key_id", None) and self.opts.get("email", None):
            LOG.warning("The user and api key configuration settings are both enabled. Credentials will default to the "
                        "api key settings.")
        LOG.info("Resilient org: %s", self.opts.get("org"))
        LOG.info("Logging Level: %s", self.opts.get("loglevel"))
        if self.opts.get("test_actions", False):
            # Make all components aware that we are in test mode
            ResilientComponent.test_mode = True

        # Make all components aware that we are in selftest mode
        ResilientComponent.IS_SELFTEST = self.IS_SELFTEST

        # Connect to events from Action Module.
        # Note: this must be done before components are loaded, because it uses
        # each component's "channel" to initiate subscription to the message queue.
        if self.opts.get("resilient", None) and self.opts["resilient"].get("actions_component", None):
            # user specified actions_component to use
            path, filename = os.path.split(self.opts["resilient"]["actions_component"])
            LOG.info("User specified actions_component: {}, from {}".format(filename, path))
            loader = Loader(init_kwargs={"opts": self.opts},
                            paths=[path])
            self.action_component = loader.load(filename)
        else:
            self.action_component = Actions(self.opts)
        self.action_component.register(self)

        # Register a `loader` to dynamically load
        # all Circuits components in the 'componentsdir' directory
        if self.auto_load_components:
            LOG.info("Components auto-load directory: %s",
                     self.opts["componentsdir"] or "(none)")
            if not self.component_loader:
                self.component_loader = ComponentLoader(self.opts)
            else:
                LOG.info("Updating and re-registering ComponentLoader")
                self.component_loader.opts = self.opts
            self.component_loader.register(self)

    def config_logging(self, logdir, loglevel, logfile, log_max_bytes, log_backup_count):
        """ set up some logging """
        global LOG_PATH, LOG, logging_initialized

        LOG_PATH = os.path.join(logdir, logfile)
        LOG_PATH = os.path.expanduser(LOG_PATH)
        LOG = logging.getLogger(__name__)

        # Only do this once! (mostly for pytest)
        if logging_initialized:
            LOG.info("Logging already initialized.")
            return
        logging_initialized = True

        # Ignore syslog errors from message-too-long
        logging.raiseExceptions = False

        try:
            numeric_level = getattr(logging, loglevel)
            logging.getLogger().setLevel(numeric_level)
        except AttributeError as e:
            logging.getLogger().setLevel(logging.INFO)
            LOG.warning("Invalid logging level specified. Using INFO level")

        LOG.addFilter(RedactingFilter())

        file_handler = RotatingFileHandler(LOG_PATH, maxBytes=log_max_bytes,
                                           backupCount=log_backup_count)
        file_handler.setFormatter(logging.Formatter(self.FILE_LOG_FORMAT))
        file_handler.addFilter(RedactingFilter())
        logging.getLogger().addHandler(file_handler)

        syslog = logging.handlers.SysLogHandler()
        syslog.setFormatter(logging.Formatter(self.SYSLOG_LOG_FORMAT))
        syslog.addFilter(RedactingFilter())
        logging.getLogger().addHandler(syslog)

        stderr = logging.StreamHandler()
        stderr.setFormatter(logging.Formatter(self.STDERR_LOG_FORMAT))
        stderr.addFilter(RedactingFilter())
        logging.getLogger().addHandler(stderr)

        if LOG.getEffectiveLevel() == logging.DEBUG:
            self += Debugger(logger=LOG)

    def load_all_success(self, event):
        """OK, component loader says we're ready"""
        LOG.info("Components loaded")

        # For debugging, print out the tree of all loaded components
        def walk(depth, component):
            yield (u"  " * depth) + repr(component)
            if isinstance(component, BaseComponent) and callable(component.events):
                for event in component.events():
                    channels = ", ".join([h.channel or '*' for h in list(component._handlers[event])])
                    yield u"{}{}/{}".format((u"  " * (depth+1)), event, channels)
            if isinstance(component, Manager):
                for c in component.components:
                    for thing in walk(depth+1, c):
                        yield thing
        tree = walk(1, self.root)
        LOG.debug(u"Components:\n" + ("\n".join(tree)))

    def load_all_failure(self, event):
        """OK, component loader says we're unable to start"""
        LOG.error("A component failed to load.  The application cannot start.")
        self.stop()

    def reload_opts(self):
        """Reload the configuration options in case they changed"""
        LOG.debug("Reload opts")
        self.opts.update(AppArgumentParser().parse_args())

    def started(self, event, component):
        """Started Event Handler"""
        LOG.info("App Started")

    def stopped(self, event, component):
        """Stopped Event Handler"""
        LOG.info("App Stopped")


def get_lock():
    """Create a filelock"""

    # The run() method uses a file lock in the user's ~/.resilient directory to prevent multiple instances
    # of resilient circuits running.  You can override the lockfile name in the
    # (and so allow multiple) by setting APP_LOCK_FILE in the environment.
    app_lock_file = os.environ.get("APP_LOCK_FILE", "")

    if not app_lock_file:
        lockfile = os.path.expanduser(os.path.join("~", ".resilient", "resilient_circuits_lockfile"))
        resilient_dir = os.path.dirname(lockfile)
        if not os.path.exists(resilient_dir):
            os.makedirs(resilient_dir)
    else:
        lockfile = os.path.expanduser(app_lock_file)
    lock = filelock.FileLock(lockfile)
    return lock


def run(*args, **kwargs):
    """Main app"""

    # define lock
    # this prevents multiple, identical circuits from running at the same time
    lock = get_lock()

    # The main app component initializes the Resilient services
    global application
    try:
        # attempt to lock file, wait 1 second for lock
        with lock.acquire(timeout=1):
            assert lock.is_locked
            application = App(*args, **kwargs)
            application.run()

    except filelock.Timeout:
        # file is probably already locked
        print("Failed to acquire lock on {0} - "
              "you may have another instance of Resilient Circuits running".format(os.path.abspath(lock.lock_file)))
    except OSError as exc:
        # Some other problem accessing the lockfile
        print("Unable to lock {0}: {1}".format(os.path.abspath(lock.lock_file), exc))
    # finally:
    #    LOG.info("App finished.")


if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

from shutil import rmtree
import pytest
from mock import patch
from resilient import get_config_file
from resilient_circuits.bin.resilient_circuits_cmd import *


DEFAULT_CONFIG_FILENAME = "app.config"
DEFAULT_CONFIG_FILE = os.path.expanduser(
    os.path.join("~", ".resilient", DEFAULT_CONFIG_FILENAME))


def mock_os_path_exists(path):
    return True


def mock_makedirs(dir):
    pass


class TestGetConfigGenerate(object):

    @pytest.fixture(scope='session')
    def tmp_config_file(self, tmpdir_factory):
        tmp_config_dir = tmpdir_factory.mktemp(".resilient")
        tmp_config_file = os.path.expanduser(os.path.join(
            str(tmp_config_dir), DEFAULT_CONFIG_FILENAME))
        yield tmp_config_file
        rmtree(str(tmp_config_dir))

    def test_generate_with_env_variable(self, monkeypatch, tmp_config_file):
        """
        Test generate with environment variable $APP_CONFIG_FILE.
        """
        monkeypatch.setenv("APP_CONFIG_FILE", str(tmp_config_file))
        result = get_config_file(generate_filename=True)
        assert result == tmp_config_file

    def test_generate_with_filename(self):
        """
        Test generate with filename specified.
        """
        test_file_name = "test_file"
        result = get_config_file(
            filename=test_file_name, generate_filename=True)
        assert result == test_file_name

    @patch('resilient.co3.os.path.exists', side_effect=mock_os_path_exists)
    def test_generate_with_app_config_path_exists(self, mock_os_path_exists):
        """
        Test with default and file exists.
        """
        result = get_config_file(generate_filename=True)
        assert result == DEFAULT_CONFIG_FILE

    @patch('resilient.co3.os.makedirs', side_effect=mock_makedirs)
    def test_generate_with_dot_make_dir(self, mock_os_makedirs):
        """
        Test app.config in "~/.resilient dir with make dir.
        """
        result = get_config_file(generate_filename=True)
        assert result == DEFAULT_CONFIG_FILE



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

import time

import pkg_resources
import pytest
from resilient_circuits import ResilientComponent, constants, function, helpers
from resilient_circuits.stomp_events import HeartbeatTimeout

from tests import (AppFunctionMockComponent, MockInboundAppComponent,
                   mock_constants)
from tests.shared_mock_data import mock_paths

resilient_mock = mock_constants.RESILIENT_MOCK
config_data = mock_constants.CONFIG_DATA


def test_get_fn_names():

    class FunctionComponentA(ResilientComponent):
        @function("mock_fn")
        def _mock_function(self):
            return True

    assert helpers.get_fn_names(FunctionComponentA) == ["mock_fn"]

    class FunctionComponentB(ResilientComponent):
        @function("mock_fn_2a")
        def _other_name_a(self):
            return True

        @function("mock_fn_2b")
        def _other_name_b(self):
            return True

    assert helpers.get_fn_names(FunctionComponentB) == ["mock_fn_2a", "mock_fn_2b"]

    with pytest.raises(ValueError, match=r"Usage: @function\(api_name\)"):
        class FunctionComponentC(ResilientComponent):
            @function("mock_fn_3a", "mock_fn_3b")
            def _other_name_a(self):
                return True


def test_get_handlers_inbound_handler():
    mock_cmp_class = MockInboundAppComponent(mock_constants.MOCK_OPTS)
    mock_handlers = helpers.get_handlers(mock_cmp_class, "inbound_handler")
    assert isinstance(mock_handlers, list)
    assert mock_handlers[0][0] == "_inbound_app_mock_one"
    assert mock_handlers[0][1].channel == "{0}.{1}".format(constants.INBOUND_MSG_DEST_PREFIX, "mock_inbound_q")


def test_check_exists():
    assert helpers.check_exists("mock", {"mock": "data"}) == "data"
    assert helpers.check_exists("mock", {}) is False
    assert helpers.check_exists("mock", None) is False
    with pytest.raises(AssertionError):
        helpers.check_exists("mock", "abc")


def test_get_configs(fx_clear_cmd_line_args):
    configs = helpers.get_configs(path_config_file=mock_paths.MOCK_APP_CONFIG)
    assert isinstance(configs, dict)
    assert configs.get("host") == "resilient"


def test_validate_configs():
    mock_configs = {
        "mock_config_1": "unicode å æ ç è é unicode",
        "mock_config_2": 11,
        "mock_config_3": "<mock_config_3_here>",
        "mock_config_4": ""
    }

    mock_required_config = {
        "required": True
    }

    mock_config_2 = {
        "required": True,
        "valid_condition": lambda x: True if x >= 1 and x <= 10 else False,
        "invalid_msg": "mock_config_2 must be in range"
    }

    mock_config_3 = {
        "required": True,
        "placeholder_value": "<mock_config_3_here>"
    }

    # test unicode
    helpers.validate_configs(mock_configs, {"mock_config_1": mock_required_config})

    # test required
    with pytest.raises(ValueError, match=r"'mock_config_5' is mandatory and is not set in the config file"):
        helpers.validate_configs(mock_configs, {"mock_config_5": mock_required_config})

    # test empty string
    with pytest.raises(ValueError, match=r"'mock_config_4' is mandatory and is not set in the config file"):
        helpers.validate_configs(mock_configs, {"mock_config_4": mock_required_config})

    # test placeholder_value
    with pytest.raises(ValueError, match=r"'mock_config_3' is mandatory and still has its placeholder value of '<mock_config_3_here>' in the config file"):
        helpers.validate_configs(mock_configs, {"mock_config_3": mock_config_3})

    # test valid_condition fails
    with pytest.raises(ValueError, match=r"mock_config_2 must be in range"):
        helpers.validate_configs(mock_configs, {"mock_config_2": mock_config_2})

    # test valid_condition passes
    mock_configs["mock_config_2"] = 5
    helpers.validate_configs(mock_configs, {"mock_config_2": mock_config_2})


def test_get_packages():

    pkgs = helpers.get_packages(pkg_resources.working_set)

    for pkg in pkgs:
        assert len(pkg) == 2
        assert isinstance(pkg[0], str)
        assert isinstance(pkg[1], str)


def test_get_packages_not_a_working_set_obj():

    invalid_ws = "this is not a working set"

    pkgs = helpers.get_packages(invalid_ws)

    assert pkgs == invalid_ws


def test_env_str():

    env_str = helpers.get_env_str(pkg_resources.working_set)

    assert "Environment" in env_str
    assert "Python Version" in env_str
    assert "Installed packages" in env_str
    assert "\n\tresilient-circuits" in env_str


def test_env_str_with_env_var(fx_add_proxy_env_var):

    env_str = helpers.get_env_str(pkg_resources.working_set)

    assert "Environment" in env_str
    assert "Python Version" in env_str
    assert "Installed packages" in env_str
    assert "\n\tresilient-circuits" in env_str
    assert "Connecting through proxy: 'https://192.168.0.5:3128'" in env_str


def test_get_queue(caplog):
    assert helpers.get_queue("/queue/actions.201.fn_main_mock_integration") == ("actions", "201", "fn_main_mock_integration")
    assert helpers.get_queue("/queue/inbound_destination.111.inbound_app_mock") == ("inbound_destination", "111", "inbound_app_mock")
    assert helpers.get_queue("inbound_destination.111.inbound_app_mock") == ("inbound_destination", "111", "inbound_app_mock")
    assert helpers.get_queue("111.inbound_app_mock") is None
    assert helpers.get_queue("") is None
    assert helpers.get_queue(None) is None
    assert "Could not get queue name" in caplog.text


class TestIsASelftestActionsComponent:
    @pytest.mark.parametrize("circuits_app", [{"IS_SELFTEST": True}], indirect=True)
    def test_is_this_a_selftest_action_component(self, circuits_app):
        assert helpers.is_this_a_selftest(circuits_app.app.action_component) is True


class TestIsNotASelftestActionsComponent:
    @pytest.mark.parametrize("circuits_app", [{"IS_SELFTEST": False}], indirect=True)
    def test_is_this_a_selftest_action_component(self, circuits_app):
        assert helpers.is_this_a_selftest(circuits_app.app.action_component) is False


class TestIsASelftestResilientComponent:
    @pytest.mark.parametrize("circuits_app", [{"IS_SELFTEST": True}], indirect=True)
    def test_is_this_a_selftest_base_component(self, circuits_app):
        c = AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app)
        assert helpers.is_this_a_selftest(c) is True


class TestIsNotASelftestResilientComponent:
    @pytest.mark.parametrize("circuits_app", [{"IS_SELFTEST": False}], indirect=True)
    def test_is_this_a_selftest_base_component(self, circuits_app):
        c = AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app)
        assert helpers.is_this_a_selftest(c) is False


def test_should_timeout():
    start_time = time.time()
    time.sleep(2)
    assert helpers.should_timeout(start_time, 1) is True


def test_should_not_timeout():
    start_time = time.time()
    assert helpers.should_timeout(start_time, 10) is False


def test_get_usr():
    assert helpers.get_user({"api_key_id": "abc", "email": None}) == "abc"
    assert helpers.get_user({"api_key_id": None, "email": "def"}) == "def"
    assert helpers.get_user({"api_key_id": "", "email": ""}) is None
    assert helpers.get_user({"api_key_id": None, "email": None}) is None


def test_filter_heartbeat_timeout_events():
    hb1 = HeartbeatTimeout(10)
    hb2 = HeartbeatTimeout(20)
    hb3 = HeartbeatTimeout(30)
    hb4 = HeartbeatTimeout()

    heartbeat_timeouts = [hb2, hb4, hb3, hb1]

    filtered_hbs = helpers.filter_heartbeat_timeout_events(heartbeat_timeouts)

    for hb in filtered_hbs:
        assert hb.ts != -1

    assert filtered_hbs[0].ts == 10
    assert filtered_hbs[1].ts == 20
    assert filtered_hbs[-1].ts == 30


def test_filter_heartbeat_timeout_events_empty():
    assert helpers.filter_heartbeat_timeout_events([]) == []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

import pytest
from resilient_circuits import (ResilientComponent, app_function, constants,
                                inbound_app)
from resilient_circuits.action_message import FunctionException_
from resilient_lib import IntegrationError

from tests import (AppFunctionMockComponent, MockFunctionComponent,
                   MockInboundAppComponent, helpers, mock_constants)

resilient_mock = mock_constants.RESILIENT_MOCK
config_data = mock_constants.CONFIG_DATA


class TestFunctionDecorator:

    def test_runs(self, circuits_app):
        MockFunctionComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        results = helpers.call_app_function(mock_constants.MOCK_FN_NAME_ONE, {"input_one": "abc"}, circuits_app)
        assert results.get("content").get("malware") is True

    def test_handles_StatusMessage(self, circuits_app):
        MockFunctionComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        mock_status_message = helpers.call_app_function(mock_constants.MOCK_FN_NAME_ONE, {"input_one": "abc"}, circuits_app, status_message_only=True)
        assert mock_status_message.text == u"Mock զ է ը թ ժ ի լ StatusMessage 1"

    def test_fn_trap_exception_False(self, circuits_app):
        MockFunctionComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)

        with pytest.raises(FunctionException_, match=r"mock error message with unicode"):
            helpers.call_app_function(mock_constants.MOCK_FN_NAME_EX, {"input_one": "abc"}, circuits_app)


class TestInboundAppDecorator:

    def test_basic_decoration(self):
        assert MockInboundAppComponent.inbound_app_mock.handler is True
        assert MockInboundAppComponent.inbound_app_mock.inbound_handler is True
        assert MockInboundAppComponent.inbound_app_mock.names == (mock_constants.MOCK_INBOUND_Q_NAME,)
        assert MockInboundAppComponent.inbound_app_mock.priority == 0
        assert MockInboundAppComponent.inbound_app_mock.channel == "{0}.{1}".format(constants.INBOUND_MSG_DEST_PREFIX, mock_constants.MOCK_INBOUND_Q_NAME)
        assert MockInboundAppComponent.inbound_app_mock.override is False
        assert MockInboundAppComponent.inbound_app_mock.event is True

    def test_inbound_app_mock_runs(self, circuits_app):
        MockInboundAppComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        event_args = helpers.call_inbound_app(circuits_app, mock_constants.MOCK_INBOUND_Q_NAME_CREATE)
        assert event_args[1] == [u"Mock incident created with unicode զ է ը թ"]

    def test_inbound_app_mock_handles_Exception(self, circuits_app):
        MockInboundAppComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        with pytest.raises(IntegrationError, match=r"mock error message with unicode"):
            helpers.call_inbound_app(circuits_app, mock_constants.MOCK_INBOUND_Q_NAME_EX)

    def test_too_many_q_names(self):
        with pytest.raises(ValueError, match=r"Usage: @inbound_app\(<inbound_destination_api_name>\)"):
            class MockInboundAppComponent2(ResilientComponent):
                @inbound_app("mock_q_2", "mock_q_3")
                def inbound_app_mock_2(self, message, *args, **kwargs):
                    return


class TestAppFunctionDecorator:

    def test_basic_decoration(self):
        assert AppFunctionMockComponent._app_function_mock_one.handler is True
        assert AppFunctionMockComponent._app_function_mock_one.function is True
        assert AppFunctionMockComponent._app_function_mock_one.names == (mock_constants.MOCK_APP_FN_NAME_ONE,)
        assert AppFunctionMockComponent._app_function_mock_one.priority == 0
        assert AppFunctionMockComponent._app_function_mock_one.channel == u"functions." + mock_constants.MOCK_APP_FN_NAME_ONE
        assert AppFunctionMockComponent._app_function_mock_one.override is False
        assert AppFunctionMockComponent._app_function_mock_one.event is True

    def test_runs(self, circuits_app):
        AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        helpers.call_app_function(mock_constants.MOCK_APP_FN_NAME_ONE, {"input_one": "abc"}, circuits_app)

    def test_handles_FunctionResult(self, circuits_app):
        mock_fn_inputs = {"input_one": u"abc", "input_two": u"unicode ઠ ડ ઢ ણ ત થ દ ધ ન પ ફ input"}
        AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        mock_result_obj = helpers.call_app_function(mock_constants.MOCK_APP_FN_NAME_ONE, mock_fn_inputs, circuits_app, full_result_obj=True)

        assert mock_result_obj.name == "app_function_mock_one"
        assert mock_result_obj.success is True
        assert mock_result_obj.reason is None

        mock_results = mock_result_obj.value

        assert mock_results["version"] == 2.0
        assert mock_results["success"] is True
        assert mock_results["reason"] is None
        assert mock_results["inputs"]["input_one"] == mock_fn_inputs["input_one"]
        assert mock_results["inputs"]["input_two"] == mock_fn_inputs["input_two"]
        assert mock_results["content"]["malware"] is True

    def test_handles_StatusMessage(self, circuits_app):
        AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        mock_status_message = helpers.call_app_function(mock_constants.MOCK_APP_FN_NAME_ONE, {"input_one": "abc"}, circuits_app, status_message_only=True)
        assert mock_status_message.text == u"Mock զ է ը թ ժ ի լ StatusMessage 1"

    def test_app_trap_exception_False(self, circuits_app):
        AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
        with pytest.raises(IntegrationError, match=r"mock error message with unicode"):
            helpers.call_app_function(mock_constants.MOCK_APP_FN_NAME_EX, {"input_one": "abc"}, circuits_app)

    def test_too_many_function_names(self):
        with pytest.raises(ValueError, match=r"Usage: @app_function\(api_name\)"):
            class AppFunctionMockComponent2(ResilientComponent):
                @app_function("mock_function_2", "mock_function_3")
                def mock_function_2(self, fn_inputs, **kwargs):
                    return


def test_fn_trap_exception_True(circuits_app):
    circuits_app.app.opts[constants.APP_CONFIG_TRAP_EXCEPTION] = True
    MockFunctionComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app.component_loader)
    results = helpers.call_app_function(mock_constants.MOCK_FN_NAME_EX, {"input_one": "abc"}, circuits_app)
    assert results.get("success") is False
    assert "mock error message with unicode" in results.get("reason")


def test_app_trap_exception_False(circuits_app):
    circuits_app.app.opts[constants.APP_CONFIG_TRAP_EXCEPTION] = False
    AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app)

    with pytest.raises(IntegrationError, match=r"mock error message with unicode"):
        helpers.call_app_function(mock_constants.MOCK_APP_FN_NAME_EX, {"input_one": "abc"}, circuits_app)


def test_app_custom_result_True(circuits_app):
    circuits_app.app.opts[constants.APP_CONFIG_TRAP_EXCEPTION] = False
    AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS).register(circuits_app.app)
    results = helpers.call_app_function(mock_constants.MOCK_APP_FN_NAME_CUSTOM_RESULT, {"input_one": "abc"}, circuits_app)
    assert results.get("custom_key", "") == "custom_value"



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Test Server component for injecting test Action Module messages"""

from __future__ import print_function
import logging
import time
import json
import struct
import uuid
from circuits import Component, Event, handler
from circuits.net.sockets import TCPServer
from circuits.net.events import write
from resilient_circuits.action_message import ActionMessage, FunctionMessage, InboundMessage
from resilient_circuits import constants, helpers


DEFAULT_TEST_HOST = "localhost"
DEFAULT_TEST_PORT = 8008
DEFAULT_WORKFLOW_INSTANCE_ID = 999

LOG = logging.getLogger(__name__)


class SubmitTestAction(Event):
    """ Circuits event to insert a test Action Message """
    def __init__(self, queue, msg_id, message):
        if not all((queue, msg_id, message)):
            raise ValueError("queue, msg_id, and message are required")
        super(SubmitTestAction, self).__init__(queue=queue, msg_id=msg_id, message=message)


class SubmitTestFunction(Event):
    """ Circuits event to insert a test Function Message """
    def __init__(self, function_name, function_params):
        if not function_name or not isinstance(function_params, dict):
            raise ValueError("function_name and function_params are required")
        msg_id = str(uuid.uuid4())
        super(SubmitTestFunction, self).__init__(queue="example", msg_id=msg_id, message={
            "function": {
                "name": function_name
            },
            "inputs": function_params,
            "workflow_instance": {
                "workflow_instance_id": DEFAULT_WORKFLOW_INSTANCE_ID
            }
        })


class SubmitTestInboundApp(Event):

    def __init__(self, queue_name, action, content, message=None):

        msg_id = str(uuid.uuid4())

        if not message:
            message = {
                "origin": {
                    "source": "mock_source",
                    "name": "mock.domain.com",
                    "id": "",
                    "version": "1.1.1"
                },
                "schema_version": "1.0",
                "content_type": "offense",
                "content_date": "2020-11-15T00:00:00+0600",
                "content_schema_version": "1.0",
                "action": action,
                "content": content
            }

        super(SubmitTestInboundApp, self).__init__(queue_name=queue_name, msg_id=msg_id, message=message)


class ResilientTestActions(Component):
    """ Mock the stomp connection for testing"""

    def __init__(self, org_id, host=DEFAULT_TEST_HOST, port=DEFAULT_TEST_PORT):
        super(ResilientTestActions, self).__init__()
        self.org_id = org_id
        bind = (host, port)
        LOG.debug("Binding test server to %s:%d", host, port)
        TCPServer(bind).register(self)
        self.messages_in_progress = {}
        self.actions_sent = {}

    def usage(self):
        return "Submit actions with format: <queue> <message json>"

    @handler("SubmitTestAction", "SubmitTestFunction")
    def _submit_message(self, event, queue, msg_id, message, channel="*"):
        """ Create and fire an ActionMessage """
        try:
            message_id = "ID:resilient-54199-{val}-6:2:12:1:1".format(val=msg_id)
            reply_to = "/queue/acks.{org}.{queue}".format(org=self.org_id,
                                                          queue=queue)
            destination = "/queue/actions.{org}.{queue}".format(org=self.org_id,
                                                                queue=queue)
            headers = {"reply-to": reply_to,
                       "expires": "0",
                       "timestamp": str(int(time.time()) * 1000),
                       "destination": destination,
                       "correlation-id": "invid:390",
                       "persistent": "True",
                       "priority": "4",
                       "Co3MessagePayload": "ActionDataDTO",
                       "Co3ContentType": "application/json",
                       "message-id": message_id,
                       "Co3ContextToken": "dummy",
                       "subscription": "stomp-{queue}".format(queue=queue)}
            try:
                sock = self.actions_sent.get(msg_id)
                if not isinstance(message, dict):
                    message = json.loads(message)
                    assert(isinstance(message, dict))
            except Exception as e:
                LOG.exception("Bad Message<action %d>: %s", msg_id, message)
                if sock:
                    msg = "Bad Message<action %d>! %s" % (msg_id, str(e))
                    self.fire_message(sock, msg)
                return

            if message.get("function"):
                channel = "functions." + message["function"]["name"]
                action_event = FunctionMessage(source=self.parent,
                                               headers=headers,
                                               message=message,
                                               test=True,
                                               test_msg_id=msg_id)
            else:
                channel = "actions." + queue
                action_event = ActionMessage(source=self.parent,
                                             headers=headers,
                                             message=message,
                                             test=True,
                                             test_msg_id=msg_id)
            action_event.parent = event
            self.fire(action_event, channel)
            if sock:
                self.fire_message(sock, "Action Submitted<action %d>" % msg_id)
        except Exception as e:
            LOG.exception("Action Failed<action %d>", msg_id)
            if sock:
                msg = "Action Failed<action %d>: %s" % (msg_id,
                                                        str(e))
                self.fire_message(sock, msg)

    @handler("SubmitTestInboundApp")
    def _submit_inbound_message(self, event, queue_name, msg_id, message, channel="*"):
        try:
            message_id = "ID:resilient-54199-{msg_id}-6:2:12:1:1".format(msg_id=msg_id)
            subscription = "{0}.{1}.{2}".format(constants.INBOUND_MSG_DEST_PREFIX, self.org_id, queue_name)
            destination = "/queue/{0}".format(subscription)
            queue = helpers.get_queue(destination)

            headers = {
                "timestamp": str(int(time.time()) * 1000),
                "persistent": "True",
                "message-id": message_id,
                "priority": "4",
                "subscription": subscription,
                "destination": destination,
                "expires": "0"
            }

            try:

                if not isinstance(message, dict):
                    message = json.loads(message)
                    assert(isinstance(message, dict))

            except Exception as e:
                LOG.exception("Bad InboundMessage <message_id: %s>: %s", msg_id, message)
                raise e

            if queue and queue[0] == constants.INBOUND_MSG_DEST_PREFIX:
                channel = u"{0}.{1}".format(constants.INBOUND_MSG_DEST_PREFIX, queue_name)
                action_event = InboundMessage(source=self.parent,
                                              queue=queue,
                                              headers=headers,
                                              message=message,
                                              test=True,
                                              test_msg_id=msg_id)

            action_event.parent = event
            self.fire(action_event, channel)

        except Exception as e:
            LOG.exception("Failed to SubmitTestInboundApp <message_id: %s>", msg_id)
            raise e

    @handler("read")
    def process_data(self, sock, data):
        """ Process data received from TCP stream """
        msg_id = -1  # default
        try:
            data_as_bytes = bytearray(data)
            if sock not in self.messages_in_progress:
                # New message, parse size out
                msg_size = struct.unpack('>I', data_as_bytes[:4])[0]
                partial_msg = data_as_bytes[4:msg_size + 4]
            else:
                # Retrieve what we had so far
                msg_size, partial_msg = self.messages_in_progress.pop(sock)
                partial_msg = partial_msg + data_as_bytes

            if len(partial_msg) < msg_size:
                # Store what we have so far and wait for more
                self.messages_in_progress[sock] = msg_size, partial_msg
                return
            else:
                # Complete message, parse out msg_id, queue, and msg
                msg_id = struct.unpack('>I', partial_msg[:4])[0]
                data = partial_msg[4:msg_size]
                remainder = partial_msg[msg_size:]
                data = data.strip().decode('utf-8')
                if ' ' not in data or len(data) < 2:
                    # Bad Command
                    msg = ("<action %d>: " % msg_id) + self.usage()
                    self.fire_message(sock, msg)
                else:
                    queue, message = data.split(' ', 1)
                    self.actions_sent[msg_id] = sock
                    self.fire(SubmitTestAction(queue, msg_id, message))

                if remainder:
                    # Process remainder
                    self.process_data(sock, remainder)
        except Exception as e:
            LOG.exception("Action Failed<action %d>", msg_id)
            msg = "Action Failed<action %d>: %s" % (msg_id, str(e))
            self.fire_message(sock, msg)

    def connect(self, sock, host, port):
        """Triggered for new connecting TCP clients"""
        pass

    def test_response(self, msg_id, message):
        """ If action orginated from test client, send response"""
        LOG.debug("Received message for test client")
        sock = self.actions_sent.get(msg_id)
        if sock:
            msg = "RESPONSE<action %d>: %s" % (msg_id, message)
            self.fire_message(sock, msg)

    def done(self, event):
        """Handler when done"""
        status = yield self.wait(event)
        status = status.value
        sock = self.actions_sent.get(event.test_msg_id)
        if isinstance(status, Exception):
            msg = "Action Failed<action %d>: %s" % (event.test_msg_id,
                                                    str(Exception))
            self.fire_message(sock, msg)
            raise status
        else:
            status = status + "\n"
            msg = "Action Completed<action %d>: %s" % (event.test_msg_id,
                                                       status)
            self.fire_message(sock, msg)

    def fire_message(self, sock, message):
        """ Prefix message with length before firing"""
        message = message.encode()
        message = struct.pack('>I', len(message)) + message
        LOG.debug("Firing Message: %s", message)
        self.fire(write(sock, message))



#
from .shared_mock_data import mock_constants, mock_paths
from .shared_mock_data.mock_component import MockInboundAppComponent
from .shared_mock_data.mock_app_function_component import AppFunctionMockComponent
from .shared_mock_data.mock_function_component import MockFunctionComponent



# -*- coding: utf-8 -*-
# pragma pylint: disable=unused-argument, no-self-use
"""Function implementation"""

import logging
from resilient_circuits import ResilientComponent, handler, inbound_app
from resilient_lib import IntegrationError
from tests import mock_constants


LOG = logging.getLogger(__name__)
PACKAGE_NAME = "fn_main_mock_integration"
QUEUE_NAME = "mock_inbound_q"


class MockInboundAppComponent(ResilientComponent):
    """Component that implements Resilient function 'mock_function_one''"""

    def __init__(self, opts):
        """constructor provides access to the configuration options"""
        super(MockInboundAppComponent, self).__init__(opts)
        self.app_configs = opts.get(PACKAGE_NAME, {})

    @handler("reload")
    def _reload(self, event, opts):
        """Configuration options have changed, save new values"""
        self.app_configs = opts.get(PACKAGE_NAME, {})

    @inbound_app(QUEUE_NAME)
    def _inbound_app_mock_one(self, message, inbound_action):

        if inbound_action == "create":
            msg_content = message.get("content", {})
            LOG.info(u"Creating incident\nIncident Description: %s", msg_content.get("description", "None"))

        elif inbound_action == "close":
            LOG.info(u"Closing incident")

        elif inbound_action == "update":
            LOG.info(u"Updating incident")

        elif inbound_action == "artifact_added":
            LOG.info(u"Updating incident")

        else:
            LOG.error(u"Unsupported functionality. Message: %s", message)

        yield "Done!"

    @inbound_app(mock_constants.MOCK_INBOUND_Q_NAME)
    def inbound_app_mock(self, message, inbound_action):
        assert isinstance(message, dict)
        yield inbound_action

    @inbound_app(mock_constants.MOCK_INBOUND_Q_NAME_CREATE)
    def inbound_app_mock_create(self, message, inbound_action):
        assert inbound_action == "create"
        assert isinstance(message, dict)
        yield u"Mock incident created with unicode զ է ը թ"

    @inbound_app(mock_constants.MOCK_INBOUND_Q_NAME_EX)
    def inbound_app_mock_raise_exception(self, message, inbound_action):
        raise IntegrationError(u"mock error message with unicode զ է ը թ ժ ի լ խ")



# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.
""" stompest StompFrameTransport allowing for ssl.wrap_socket """

import logging
import ssl
import socket
from stompest.sync.transport import StompFrameTransport
from stompest.error import StompConnectionError
import socks
from resilient import constants, helpers

LOG = logging.getLogger(__name__)


class EnhancedStompFrameTransport(StompFrameTransport):
    """ add support for older ssl module and http proxy """

    proxy_host = None
    proxy_port = None
    proxy_user = None
    proxy_password = None

    @staticmethod
    def match_hostname(cert, hostname):
        """ Check that hostname matches cert """
        names = []
        # Python 3 has an ssl.match_hostname method, which does hostname validation.
        try:
            ssl.match_hostname(cert, hostname)
            return
        except AttributeError as err:
            # We don't have the backported python 3 ssl module, do a simplified check
            for sub in cert.get('subject', ()):
                for key, value in sub:
                    if key == 'commonName':
                        names.append(value)
                        if value == hostname:
                            return
        raise Exception("{0} does not match the expected value in the certificate {1}".format(hostname, str(names)))

    def connect(self, timeout=None):
        """ Allow older versions of ssl module, allow http proxy connections """
        LOG.debug("stomp_transport.connect()")
        ssl_params = None
        if isinstance(self.sslContext, dict):
            # This is actually a dictionary of ssl parameters for wrapping the socket
            ssl_params = self.sslContext
            self.sslContext = None

        proxy_details = helpers.get_and_parse_proxy_env_var(constants.ENV_HTTPS_PROXY)
        proxy_type = socks.HTTP

        if helpers.is_env_proxies_set() and proxy_details:

            if helpers.is_in_no_proxy(self.host):
                self.proxy_host = None
                self.proxy_port = None
                self.proxy_user = None
                self.proxy_password = None

            else:
                self.proxy_host = proxy_details.get("hostname", "")
                self.proxy_port = proxy_details.get("port")
                self.proxy_user = proxy_details.get("username", "")
                self.proxy_password = proxy_details.get("password", "")

        try:
            if self.proxy_host:
                LOG.info("Connecting through proxy %s", self.proxy_host)
                self._socket = socks.socksocket()
                self._socket.set_proxy(proxy_type, self.proxy_host, self.proxy_port, True,
                                       username=self.proxy_user, password=self.proxy_password)
            else:
                self._socket = socket.socket()

            self._socket.settimeout(timeout)
            self._socket.connect((self.host, self.port))

            if ssl_params:
                # For cases where we don't have a modern SSLContext (so no SNI)
                cert_required = ssl.CERT_REQUIRED if ssl_params["ca_certs"] else ssl.CERT_NONE
                self._socket = ssl.wrap_socket(
                    self._socket,
                    keyfile=ssl_params['key_file'],
                    certfile=ssl_params['cert_file'],
                    cert_reqs=cert_required,
                    ca_certs=ssl_params['ca_certs'],
                    ssl_version=ssl_params['ssl_version'])
                if cert_required:
                    LOG.info("Performing manual hostname check")
                    cert = self._socket.getpeercert()
                    self.match_hostname(cert, self.host)

            if self.sslContext:
                self._socket = self.sslContext.wrap_socket(self._socket, server_hostname=self.host)

        except IOError as e:
            raise StompConnectionError('Could not establish connection [%s]' % e)
        self._parser.reset()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2020. All Rights Reserved.

"""
Shared pytest fixtures

Note:
    -   Code after the 'yield' statement in a fixture
        is ran after the test (or scope i.e. test session) has complete
    -   fx_ prefixes a 'fixture'
    -   Put fixture logic in separate 'private' function so we
        can share logic between fixtures
    -   Fixture must have BEFORE and AFTER docstring
"""

import copy
import os
import sys

import pkg_resources
import pytest
import requests_mock
from resilient import ImportDefinition
from resilient import constants as res_constants
from resilient.co3 import SimpleClient

from tests.shared_mock_data import mock_paths


@pytest.fixture
def fx_simple_client():
    """
    Before: Creates sample SimpleClient object with mocked URI https://example.com
    After: Nothing
    """
    simple_client = SimpleClient(org_name="Mock Org", base_url="https://example.com")
    requests_adapter = requests_mock.Adapter()
    simple_client.session.mount(u'https://', requests_adapter)
    simple_client.org_id = 201

    mock_inc_fields_uri = "{0}/rest/orgs/{1}/types/incident/fields".format(simple_client.base_url, simple_client.org_id)
    requests_adapter.register_uri('GET', mock_inc_fields_uri, status_code=200, json={})

    mock_action_fields_uri = "{0}/rest/orgs/{1}/types/actioninvocation/fields".format(simple_client.base_url, simple_client.org_id)
    requests_adapter.register_uri('GET', mock_action_fields_uri, status_code=200, json={})

    mock_actions_uri = "{0}/rest/orgs/{1}/actions".format(simple_client.base_url, simple_client.org_id)
    requests_adapter.register_uri('GET', mock_actions_uri, status_code=200, json={"entities": []})

    yield (simple_client, requests_adapter)


@pytest.fixture
def fx_clear_cmd_line_args():
    """
    Before: Clear command line arguments
    After: Set the cmd line args back to its original value
    """
    original_cmd_line = copy.deepcopy(sys.argv)

    sys.argv = [sys.argv[0]]

    yield

    sys.argv = original_cmd_line


@pytest.fixture
def fx_add_entry_point(ep_str, path_dist):
    """
    Before: get original pkg_resources.working_set + add Entry Point at path_dist
    After: set pkg_resources.working_set to original

    Params:
        ep_str: str e.g. "mock_component:MockInboundComponent"
        path_dist: str e.g. "user/resilient-circuits/tests/shared_mock_data"
    """
    original_working_set = copy.deepcopy(pkg_resources.working_set)

    ep_name = ep_str.split(":")[1]

    distribution = pkg_resources.Distribution(path_dist)
    ep_obj = pkg_resources.EntryPoint.parse("{0}={1}".format(ep_name, ep_str), dist=distribution)
    distribution._ep_map = {'resilient.circuits.components': {ep_name: ep_obj}}
    pkg_resources.working_set.add(distribution)

    yield

    pkg_resources.working_set = original_working_set


@pytest.fixture
def fx_add_proxy_env_var():
    """
    Before: set HTTPS_PROXY, HTTP_PROXY and NO_PROXY
    After: unset HTTPS_PROXY, HTTP_PROXY and NO_PROXY
    """
    os.environ[res_constants.ENV_HTTPS_PROXY] = "https://192.168.0.5:3128"
    os.environ[res_constants.ENV_HTTP_PROXY] = "http://192.168.0.5:3128"
    os.environ[res_constants.ENV_NO_PROXY] = "subdomain.example.com,192.168.1.5,.example.com"

    yield

    os.environ[res_constants.ENV_HTTPS_PROXY] = ""
    os.environ[res_constants.ENV_HTTP_PROXY] = ""
    os.environ[res_constants.ENV_NO_PROXY] = ""


@pytest.fixture
def fx_read_mock_definition():
    """
    Before: reads the mock_import_definition.txt file and yields an ImportDefinition of it
    After: N/A
    """
    with open(mock_paths.MOCK_IMPORT_DEFINITION, mode="r") as f:
        lines = f.readlines()
        b64 = "\n".join(lines)

    yield ImportDefinition(b64)



# -*- coding: utf-8 -*-

from resilient_circuits import (FunctionError, FunctionResult,
                                ResilientComponent, StatusMessage, function)
from resilient_lib import ResultPayload, IntegrationError
from tests import mock_constants

PACKAGE_NAME = mock_constants.MOCK_PACKAGE_NAME


class MockFunctionComponent(ResilientComponent):

    def __init__(self, opts):
        super(MockFunctionComponent, self).__init__(opts)
        self.options = opts.get(PACKAGE_NAME, {})

    @function(mock_constants.MOCK_FN_NAME_ONE)
    def _fn_mock_fn_one(self, event, *args, **kwargs):
        try:
            rp = ResultPayload(mock_constants.MOCK_FN_NAME_ONE, **kwargs)
            yield StatusMessage(u"Mock զ է ը թ ժ ի լ StatusMessage 1")
            yield StatusMessage(u"Mock StatusMessage 2")
            yield FunctionResult(rp.done(True, {"malware": True}))
        except Exception as e:
            yield FunctionError(e)

    @function(mock_constants.MOCK_FN_NAME_EX)
    def _fn_mock_raise_exception(self, event, *args, **kwargs):
        try:
            raise IntegrationError(u"mock error message with unicode զ է ը թ ժ ի լ խ")
        except Exception as e:
            yield FunctionError(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

""" Command line tool to manage and run resilient-circuits """
import argparse
import logging
import os
import sys
from collections import defaultdict
import pkg_resources
from resilient import get_config_file
from resilient_circuits import helpers, constants
from resilient_circuits.util.resilient_customize import customize_resilient
from resilient_circuits.cmds import selftest


if sys.version_info.major == 2:
    from io import open
else:
    unicode = str
try:
    # For all python < 3.2
    import backports.configparser as configparser
except ImportError:
    import configparser

try:
    from builtins import input
except ImportError:
    # Python 2
    from __builtin__ import raw_input as input

LOG = logging.getLogger(constants.CMDS_LOGGER_NAME)
LOG.setLevel(logging.INFO)
LOG.addHandler(logging.StreamHandler())


def windows_service(service_args, res_circuits_args):
    """Register a windows service"""
    try:
        import win32serviceutil
        from resilient_circuits.bin import service_wrapper

        if res_circuits_args:
            # Set the command line arguments to pass to "resilient-circuits.exe run"
            service_wrapper.irms_svc.setResilientArgs(res_circuits_args)

        sys.argv = sys.argv[0:1] + service_args
        win32serviceutil.HandleCommandLine(service_wrapper.irms_svc)
    except ImportError:
        LOG.error("Requires PYWIN32 Package. Please download and install from: "
                  "https://sourceforge.net/projects/pywin32/files/pywin32/")


def supervisor_service():
    """Register a unix service with supervisord [deprecated]"""
    pass


def manage_service(service_args, res_circuits_args):
    """Register a windows or unix service"""
    if os.name == 'nt':
        windows_service(service_args, res_circuits_args)
    else:
        LOG.error("Not implemented")
        # supervisor_service(service_args, res_circuits_args)


def run(resilient_circuits_args, restartable=False, config_file=None):
    """Run resilient-circuits"""
    # Leave only the arguments for the run command
    if restartable:
        # import here b/c it is slow
        from resilient_circuits import app_restartable as app
    else:
        from resilient_circuits import app
    sys.argv = sys.argv[0:1] + resilient_circuits_args
    kwargs = {}
    if config_file:
        kwargs = {"config_file": config_file}

    LOG.info(helpers.get_env_str(pkg_resources.working_set))

    app.run(**kwargs)


def list_installed(args):
    """print list of installed packages with their components"""
    LOG.debug("resilient-circuits.list")
    components = defaultdict(list)
    # Resilient packages can have several entry-points,
    # - resilient.circuits.config: produces default configuration sections
    # - resilient.circuits.customize: produces schema customizations
    # - resilient.circuits.components: identifies components that implement actions, functions, etc.
    # We want to list all components, but include (as "empty") packages that define the other entry-points too.
    entry_points = []
    entry_points.extend([ep for ep in pkg_resources.iter_entry_points('resilient.circuits.config')])
    entry_points.extend([ep for ep in pkg_resources.iter_entry_points('resilient.circuits.customize')])
    component_entry_points = [ep for ep in pkg_resources.iter_entry_points('resilient.circuits.components')]
    entry_points.extend(component_entry_points)
    LOG.debug(u"Found %d installed entry-points", len(entry_points))
    for ep in entry_points:
        components[ep.dist].append(ep)
    if not components:
        LOG.info(u"No resilient-circuits components are installed")
        return

    ep_list = {}
    LOG.info(u"The following packages and components are installed:")
    for dist, component_list in components.items():
        if args.verbose:
            clist = "\n\t".join([str(ep) for ep in component_list if ep in component_entry_points])
            if clist == "":
                ep = u"{} ({}):\n\t(Package does not define any components)".format(
                    dist.as_requirement(),
                    dist.egg_info)
            else:
                ep = u"{} ({}):\n\t{}".format(
                    dist.as_requirement(),
                    dist.egg_info,
                    clist)
        else:
            clist = "\n\t".join([ep.name for ep in component_list if ep in component_entry_points])
            if clist == "":
                ep = u"{}:\n\t(Package does not define any components)".format(
                    dist.as_requirement())
            else:
                ep = u"{}:\n\t{}".format(
                    dist.as_requirement(),
                    clist)

        ep_list[str(dist.as_requirement())] = ep

    for ep_item in sorted(ep_list.keys()):
        LOG.info(ep_list[ep_item])


def generate_default(install_list):
    """ return string containing entire default app.config """
    base_config_fn = pkg_resources.resource_filename("resilient_circuits", "data/app.config.base")
    entry_points = pkg_resources.iter_entry_points('resilient.circuits.configsection')
    additional_sections = []
    remaining_list = install_list[:] if install_list else []
    for entry in entry_points:
        dist = entry.dist
        package_name = entry.dist.project_name

        # if a list is provided, use it to filter which packages to add to the app.config file
        if install_list is not None and package_name not in remaining_list:
            LOG.debug("{} bypassed".format(package_name))
            continue
        elif package_name in remaining_list:
            remaining_list.remove(package_name)

        try:
            func = entry.load()
        except ImportError:
            LOG.exception(u"Failed to load configuration defaults for package '%s'", repr(dist))
            continue

        new_section = None
        try:
            config_data = func()
            if config_data:
                required_config = configparser.ConfigParser(interpolation=None)
                LOG.debug(u"Config Data String:\n%s", config_data)
                required_config.read_string(unicode(config_data))
                new_section = required_config.sections()[0]
        except:
            LOG.exception(u"Failed to get configuration defaults for package '%s'", repr(dist))
            continue

        if new_section:
            additional_sections.append(config_data)

    LOG.debug("Found %d sections to generate", len(additional_sections))

    if install_list and len(remaining_list) > 0:
        LOG.warning("%s not found. Check package name(s)", remaining_list)

    with open(base_config_fn, 'r') as base_config_file:
        base_config = base_config_file.read()
        return "\n\n".join(([base_config, ] + additional_sections))


def generate_or_update_config(args):
    """ Create or update config file based on installed components.

    :param args: Command-line args for Resilient circuits 'config' sub-command.
    """
    usage_type = "CREATING" if args.create else "UPDATING"
    # Get the config file name.
    config_filename = get_config_file(filename=args.filename, generate_filename=True)

    file_exists = os.path.exists(config_filename)

    if file_exists:
        if args.create:
            choice = ""
            while choice not in ('y', 'n'):
                choice = input(u"%s exists. Do you want to overwrite? (y/n): " % config_filename)
            if choice == 'n':
                LOG.error(u"Config file creation cancelled.")
                return
    elif args.update:
        LOG.error(u"File %s does not exist. Update cancelled.", config_filename)
        return

    LOG.info(u"%s config file %s", usage_type, config_filename)

    if args.create:
        # Write out default file
        with open(config_filename, "w+", encoding="utf-8") as config_file:
            config_file.write(generate_default(args.install_list))
            LOG.info(u"Configuration file generated: %s", config_filename)
            LOG.info(u"Please manually edit with your specific configuration values.")

    else:
        # Update existing file
        config = configparser.ConfigParser(interpolation=None)
        updated = False
        with open(config_filename, "r", encoding="utf-8") as config_file:
            first_byte = config_file.read(1)
            if first_byte != u'\ufeff':
                # Not a BOM, no need to skip first byte
                config_file.seek(0)
            config.read_file(config_file)
            existing_sections = config.sections()

        entry_points = pkg_resources.iter_entry_points('resilient.circuits.configsection')
        remaining_list = args.install_list[:] if args.install_list else []

        with open(config_filename, "a", encoding="utf-8") as config_file:
            for entry in entry_points:
                dist = entry.dist
                package_name = entry.dist.project_name
                try:
                    func = entry.load()
                except ImportError:
                    LOG.exception(u"Failed to load configuration defaults for package '%s'", repr(dist))
                    continue

                new_section = None
                try:
                    config_data = func()
                    if config_data:
                        required_config = configparser.ConfigParser(interpolation=None)
                        LOG.debug(u"Config Data String:\n%s", config_data)
                        required_config.read_string(unicode(config_data))
                        new_section = required_config.sections()[0]
                except:
                    LOG.exception(u"Failed to get configuration defaults for package '%s'", repr(dist))
                    continue

                LOG.debug(u"Required Section: %s", new_section)
                if new_section and new_section not in existing_sections:
                    if args.install_list is None or package_name in remaining_list:
                        # Add the default data for this required section to the config file
                        LOG.info(u"Adding new section '%s' for '%s'", new_section, dist)
                        if package_name in remaining_list:
                            remaining_list.remove(package_name)

                        config_file.write(u"\n" + config_data)
                        updated = True
                else:
                    LOG.debug(u"Section '%s' already present, not adding", new_section)
                    LOG.debug(u"%s %s", new_section, package_name)
                    if package_name in remaining_list:
                        remaining_list.remove(package_name)

            if args.install_list and len(remaining_list) > 0:
                LOG.warning("%s not found. Check package name(s)", remaining_list)

            if updated:
                LOG.info(u"Update finished.  "
                         u"New sections may require manual edits with your specific configuration values.")
            else:
                LOG.info(u"No updates.")


def main():
    """Main commandline"""

    parser = argparse.ArgumentParser(
        prog="resilient-circuits",
        description="Runtime environment for apps used with IBM Security SOAR",
        epilog="For support, please visit ibm.biz/soarcommunity"
    )

    parser.usage = """
    $ resilient-circuits <subcommand> ...
    $ resilient-circuits -v <subcommand> ...
    $ resilient-circuits -h
    """

    parser.add_argument("-v", "--verbose", help="Set the log level to DEBUG", action="store_true")

    subparsers = parser.add_subparsers(title="subcommands",
                                       description="one of these options must be provided",
                                       metavar="",
                                       dest="cmd")

    subparsers.required = True

    run_parser = subparsers.add_parser("run",
                                       help="Run the Resilient Circuits application")

    list_parser = subparsers.add_parser("list",
                                        help="List the installed Resilient Circuits components")

    test_parser = subparsers.add_parser("test",
                                        help="(Deprecated) Run an interactive client for testing Resilient Circuits messages")

    service_parser = subparsers.add_parser("service",
                                           help="Manage Resilient Circuits as a service")

    config_parser = subparsers.add_parser("config",
                                          help="Create or update a basic configuration file")

    customize_parser = subparsers.add_parser("customize",
                                             help="Apply customizations to the Resilient platform")

    selftest_parser = subparsers.add_parser("selftest",
                                            help="Call selftest functions for every package and print their return states")

    # Options for selftest
    selftest_parser.add_argument("-l", "--list",
                                 dest="install_list",
                                 help="Test specified list of package(s)",
                                 nargs="+")

    selftest_parser.add_argument("--print-env",
                                 help="Print Python version and list installed Packages",
                                 action="store_true")

    # Options for 'list'
    list_parser.add_argument("-v", "--verbose", action="store_true")

    # Options for 'config'
    file_option_group = config_parser.add_mutually_exclusive_group(required=True)

    file_option_group.add_argument("-u", "--update",
                                   help="Add any missing sections required for installed components",
                                   action="store_true")

    file_option_group.add_argument("-c", "--create",
                                   help="Create new config file with all required sections",
                                   action="store_true")

    config_parser.add_argument("filename",
                               help="Config file to write to; e.g. 'app.config'",
                               default="",
                               nargs="?")

    config_parser.add_argument("-l", "--list",
                               dest="install_list",
                               help="Config specified list of package(s)",
                               nargs="+")

    # Options for 'run'
    run_parser.add_argument("-r", "--auto-restart",
                            help="Automatically restart all components if config file changes",
                            action="store_true")

    run_parser.add_argument("--config-file",
                            help="Pull configuration from specified file",
                            default=None)

    run_parser.add_argument("resilient_circuits_args", help="Args to pass to app.run", nargs=argparse.REMAINDER)

    # Options for 'service'
    service_parser.add_argument("--res-circuits-args",
                                help="Arguments to pass to resilient-circuits.exe run command",
                                action="store",
                                default="")

    service_parser.add_argument("service_args", help="Args to pass to service manager", nargs=argparse.REMAINDER)

    # Options for 'customize'
    customize_parser.add_argument("-y",
                                  dest="yflag",
                                  help="Customize without prompting for confirmation",
                                  action="store_true")

    customize_parser.add_argument("-l", "--list",
                                  dest="install_list",
                                  help="Install specified list of package(s)",
                                  nargs="+")

    args, unknown_args = parser.parse_known_args()

    if args.verbose:
        LOG.debug("Verbose Logging Enabled")
        LOG.setLevel(logging.DEBUG)

    if args.cmd != "run" and len(unknown_args) > 0:
        # Shouldn't have any unknown args for other commands, generate the proper errors
        args = parser.parse_args()

    if args.cmd == "run":
        run(unknown_args + args.resilient_circuits_args,
            restartable=args.auto_restart,
            config_file=args.config_file)

    elif args.cmd == "test":
        from resilient_circuits.bin import res_action_test
        res_action_test.ResilientTestProcessor().cmdloop()

    elif args.cmd == "config":
        generate_or_update_config(args)

    elif args.cmd == "list":
        list_installed(args)

    elif args.cmd == "service":
        manage_service(unknown_args + args.service_args, args.res_circuits_args)

    elif args.cmd == "customize":
        logging.basicConfig(format='%(message)s', level=logging.INFO)
        customize_resilient(args)

    elif args.cmd == "selftest":
        selftest.execute_command(args)


if __name__ == "__main__":
    LOG.debug("CALLING MAIN")
    main()



# -*- coding: utf-8 -*

from resilient_circuits import (AppFunctionComponent, FunctionResult,
                                app_function)
from resilient_lib import IntegrationError
from tests import mock_constants

PACKAGE_NAME = mock_constants.MOCK_PACKAGE_NAME
REQUIRED_APP_CONFIGS = mock_constants.MOCK_REQUIRED_APP_CONFIGS


class AppFunctionMockComponent(AppFunctionComponent):

    def __init__(self, opts, package_name="", required_app_configs=[]):
        if not package_name:
            package_name = PACKAGE_NAME

        if not required_app_configs:
            required_app_configs = REQUIRED_APP_CONFIGS

        super(AppFunctionMockComponent, self).__init__(opts, package_name, required_app_configs)

    @app_function(mock_constants.MOCK_APP_FN_NAME_ONE)
    def _app_function_mock_one(self, fn_inputs):
        yield self.status_message(u"Mock զ է ը թ ժ ի լ StatusMessage 1")
        yield self.status_message(u"Mock StatusMessage 2")
        yield self.status_message(fn_inputs.input_one)

        fn_msg = self.get_fn_msg()

        yield self.status_message(u"Function name: {0}".format(fn_msg.get("function", {}).get("name", "Unknown")))
        yield FunctionResult({"malware": True})

    @app_function(mock_constants.MOCK_APP_FN_NAME_CUSTOM_RESULT)
    def _app_function_mock_custom_result(self, fn_inputs):
        yield FunctionResult({"custom_key": "custom_value"}, custom_results=True)

    @app_function(mock_constants.MOCK_APP_FN_NAME_EX)
    def _app_function_mock_raise_exception(self, fn_inputs):
        raise IntegrationError(u"mock error message with unicode զ է ը թ ժ ի լ խ")



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.

import pytest
from resilient_circuits.action_message import *

class TestErrors(object):
    def test_errors_subclass_value_error(self):
        assert issubclass(FunctionException_, ValueError)
        assert issubclass(FunctionError_, ValueError)

    def test_function_exception_returned_after_exception(self):
        try:
            raise ValueError("hello")
        except ValueError:
            assert issubclass(FunctionError().__class__, FunctionException_)

    def test_function_error_returned_outside_try(self):
        assert issubclass(FunctionError().__class__, FunctionError_)

    def test_function_error_returned_after_handled_exception(self):
        """
        Different behavior in Python 2.7 and 3
        In 3, after the try..catch sys.exc_info is empty, in 2.7 it's not
        """
        import sys
        PY2 = sys.version_info[0] == 2

        try:
            raise ValueError("hello")
        except ValueError:
            pass
        if PY2:
            assert issubclass(FunctionError().__class__, FunctionException_)
        else:
            assert issubclass(FunctionError().__class__, FunctionError_)

    def test_message_passed_in_error_preserved(self):
        err = FunctionError("error")
        assert str(err) == "error"

    def test_trace_is_included_unless_otherwise_said(self):
        try:
            raise ValueError("trace_included")
        except:
            err = FunctionError("message_included")
            message = str(err)
            assert "trace_included" in message

    def test_unexplicit_unicode_message_in_error(self):
        err = FunctionError("√message")
        message = str(err)
        assert "√message" in message

    def test_explicit_unicode_message_in_error(self):
        err = FunctionError(u"√message")
        message = str(err)
        assert "√message" in message

    def test_unexplicit_unicode_message_in_trace(self):
        try:
            raise ValueError("√trace")
        except ValueError:
            err = FunctionError("message_here")
            message = str(err)

            assert "√trace" in message


    def test_explicit_unicode_message_in_trace(self):
        try:
            raise ValueError(u"√trace -> weird Unicode symbol")
        except ValueError:
            err = FunctionError("message_here")
            message = str(err)

            assert message
            assert "√trace" in message


    """
    --------------------------------
    Section testing new capabilities
    --------------------------------
    """
    def test_unicode_in_message_in_trace(self):
        try:
            raise ValueError("√trace weird Unicode symbol")
        except:
            err = FunctionError("√message")
            message = str(err)
            assert "√trace" in message
            assert "√message" in message

    def test_message_passed_in_exception_preserved_no_trace(self):
        try:
            raise ValueError("value_error")
        except:
            err = FunctionError("error", trace=False)
            assert str(err) == "error"

    def test_trace_and_message_preserved(self):
        try:
            raise ValueError("trace_included")
        except:
            err = FunctionError("message_included")
            message = str(err)
            assert "trace_included" in message
            assert "message_included" in message



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Utility functions for configuration"""

from __future__ import print_function
import logging
import pkg_resources


LOG = logging.getLogger("__name__")


def get_config_data(package):
    """Read the default configuration-data section from the given package"""
    data = None
    try:
        dist = pkg_resources.get_distribution(package)
        entries = pkg_resources.get_entry_map(dist, "resilient.circuits.configsection")
        if entries:
            entry = next(iter(entries))
            func = entries[entry].load()
            data = func()
    except pkg_resources.DistributionNotFound:
        pass
    return data or ""



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

""" Circuits events for STOMP Client """

import logging
from circuits import Event
from resilient_circuits import helpers

LOG = logging.getLogger(__name__)


class StompEvent(Event):
    """A Circuits event with less verbose repr"""
    success = True

    def _repr(self):
        return ""

    def __repr__(self):
        "x.__repr__() <==> repr(x)"

        if len(self.channels) > 1:
            channels = repr(self.channels)
        elif len(self.channels) == 1:
            channels = str(self.channels[0])
        else:
            channels = ""

        data = self._repr()

        return u"<%s[%s] (%s)>" % (self.name, channels, data)


class Disconnected(StompEvent):
    def __init__(self, reconnect=True, receipt=None):
        super(Disconnected, self).__init__(receipt=receipt)
        self.reconnect = reconnect


class Disconnect(StompEvent):
    def __init__(self, reconnect=False, flush=True, receipt=None):
        super(Disconnect, self).__init__(receipt=receipt, reconnect=reconnect, flush=flush)
    pass


class Message(StompEvent):
    def __init__(self, frame):
        super(Message, self).__init__(headers=frame.headers,
                                      message=frame.body,
                                      queue=helpers.get_queue(frame.headers.get("destination", "")))
        self.frame = frame


class Send(StompEvent):
    failure = True

    def __init__(self, headers, body, destination, message_id=None):
        self.headers = headers
        self.body = body
        self.destination = destination
        self.message_id = message_id
        super(Send, self).__init__(headers=headers,
                                   body=body,
                                   destination=destination)


class ClientHeartbeat(StompEvent):
    pass


class ServerHeartbeat(StompEvent):
    pass


class Connect(StompEvent):
    def __init__(self, subscribe=False, host=None):
        super(Connect, self).__init__(host=host)
        self.subscribe = subscribe


class Connected(StompEvent):
    pass


class ConnectionFailed(StompEvent):
    reconnect = True


class OnStompError(StompEvent):
    def __init__(self, frame, err):
        headers = frame.headers if frame else {}
        body = frame.body if frame else None
        super(OnStompError, self).__init__(headers=headers,
                                           message=body,
                                           error=err)
        self.frame = frame


class HeartbeatTimeout(StompEvent):

    def __init__(self, ts=-1):
        """
        A circuits event that is fired with the current time

        :param ts: A timestamp, defaults to -1
        :type ts: int, float, optional
        """
        super(HeartbeatTimeout, self).__init__()
        self.ts = ts

    def __lt__(self, other):
        return self.ts < other.ts


class Subscribe(StompEvent):
    def __init__(self, destination, additional_headers=None, **kwargs):
        super(Subscribe, self).__init__(destination=destination, additional_headers=additional_headers, **kwargs)
        self.destination = destination


class Unsubscribe(StompEvent):
    def __init__(self, destination):
        super(Unsubscribe, self).__init__(destination=destination)
        self.destination = destination


class Ack(StompEvent):
    failure = True

    def __init__(self, frame, message_id=None):
        super(Ack, self).__init__(frame=frame)
        self.frame = frame
        self.message_id = message_id


class SelftestTerminateEvent(Event):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

"""Global accessor for the Resilient REST API"""

import logging

from resilient_circuits import constants

import resilient
from resilient import constants as res_constants

LOG = logging.getLogger(__name__)

resilient_client = None
connection_opts = None


def reset_resilient_client():
    """Reset the cached client"""
    global resilient_client
    resilient_client = None


def get_resilient_client(opts):
    """
    Get a connected instance of :class:`resilient.SimpleClient <resilient.co3.SimpleClient>`
    for the SOAR REST API

    Passes the custom header of ``Resilient-Circuits-Version: 45.0.0`` to the
    :class:`resilient.SimpleClient <resilient.co3.SimpleClient>` instantiation

    :param opts: the connection options - usually the contents of the app.config file
    :type opts: dict

    :return: a connected and verified instance of :class:`resilient.SimpleClient <resilient.co3.SimpleClient>`
    """
    global resilient_client
    global connection_opts

    new_opts = (opts.get("cafile"),
                opts.get("org"),
                opts.get("host"),
                opts.get("port"),
                opts.get("resource_prefix"),
                opts.get("proxy_host"),
                opts.get("proxy_port"),
                opts.get("proxy_user"),
                opts.get("proxy_password"),
                opts.get("email"),
                opts.get("api_key_id"))

    custom_headers = {
        constants.HEADER_CIRCUITS_VER_KEY: constants.HEADER_CIRCUITS_VER_VALUE
    }

    if new_opts != connection_opts:
        resilient_client = None
        connection_opts = new_opts

    if resilient_client:
        return resilient_client

    retry_args = {
        res_constants.APP_CONFIG_MAX_CONNECTION_RETRIES: opts.get(res_constants.APP_CONFIG_MAX_CONNECTION_RETRIES),
        res_constants.APP_CONFIG_REQUEST_MAX_RETRIES: opts.get(res_constants.APP_CONFIG_REQUEST_MAX_RETRIES),
        res_constants.APP_CONFIG_REQUEST_RETRY_DELAY: opts.get(res_constants.APP_CONFIG_REQUEST_RETRY_DELAY),
        res_constants.APP_CONFIG_REQUEST_RETRY_BACKOFF: opts.get(res_constants.APP_CONFIG_REQUEST_RETRY_BACKOFF)
    }

    resilient_client = resilient.get_client(opts, custom_headers=custom_headers, **retry_args)

    server_version = get_resilient_server_version(resilient_client)

    LOG.info("IBM Security QRadar SOAR version: v%s", server_version)

    return resilient_client

def get_resilient_server_version(res_client):
    """
    Uses get_const to get the "server_version"
    and converts it into a float of ``major.minor`` and returns it

    :param res_client: required for communication back to SOAR
    :type res_client: resilient.get_client()
    :return: the server_version in the form ``major.minor.build_number``
    :rtype: str
    """
    LOG.debug("Getting server version")
    try:
        server_version = res_client.get_const().get("server_version", {})
    except Exception:
        # fine to ignore any exceptions here as this is
        # not critical to get the server info if something goes wrong
        server_version = {}

    current_server_version = "{0}.{1}.{2}".format(server_version.get("major", 0),
            server_version.get("minor", 0), server_version.get("build_number", 0))

    return current_server_version



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

from resilient_circuits.template_functions import environment, render, render_json

def test_environment():
    env = environment()
    for custom in ['soar_datetimeformat', 'soar_substitute', 'soar_trimlist', 'soar_splitpart',
                   'pretty', 'js', 'json', 'html', 'idna', 'ps', 'iso8601',
                   'punycode', 'ldap']:
        assert custom in env.filters

def test_render():
    assert render("{{ url_data | url }}", { "url_data": "https://ibm.com?something=here"}) == "https%3A//ibm.com%3Fsomething%3Dhere"
    assert render("{{ lookup | soar_substitute('{ \"info\": \"low\"}') }}", { "lookup": "info"}) == "low"

def test_render_json():
    assert render_json('{ "date": "{{ iso8601_date | iso8601 }}" }', { "iso8601_date": 1645820357000 }) == { "date": "2022-02-25T20:19:17+00:00" }
    assert render_json('{ "date": {{ datetime | soar_datetimeformat(split_at="+") }} }', { "datetime": "2022-02-25T20:19:17+00:00" }) == { "date": 1645820357000 }


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

"""Template processing"""

# Jinja template functions
from deprecated import deprecated
from resilient_lib import global_jinja_env as lib_environment
from resilient_lib import render as lib_render
from resilient_lib import render_json as lib_render_json


@deprecated(version='44.1', reason="Use resilient-lib.global_jinja_env")
def environment():
    return lib_environment()


@deprecated(version='44.1', reason="Use resilient-lib.render")
def render(template, data):
    return lib_render(template, data)


@deprecated(version='44.1', reason="Use resilient-lib.render_json")
def render_json(template, data):
    return lib_render_json(template, data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2020. All Rights Reserved.

import copy
import sys

import pytest
from resilient_circuits.app import AppArgumentParser
from resilient_circuits.validate_configs import (MAX_NUM_WORKERS,
                                                 MIN_BACKUP_COUNT,
                                                 MIN_LOG_BYTES)
from tests.shared_mock_data import mock_paths


def test_num_workers(fx_clear_cmd_line_args):

    # Test reading from app.config
    opts = AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()
    assert isinstance(opts.get("num_workers"), int)
    assert opts.get("num_workers") == 50

    # Test default value - commented out in app.config
    opts = AppArgumentParser(config_file=mock_paths.MOCK_COMMENTED_APP_CONFIG).parse_args()
    assert isinstance(opts.get("num_workers"), int)
    assert opts.get("num_workers") == 25

    # Test overwriting
    sys.argv.extend(["--num-workers", "30"])
    opts = AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()
    assert isinstance(opts.get("num_workers"), int)
    assert opts.get("num_workers") == 30

    # Test if over limit
    sys.argv.extend(["--num-workers", str(MAX_NUM_WORKERS+1)])
    with pytest.raises(ValueError, match=r"num_workers must be in the range .*"):
        AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()


def test_global_integrations_options(fx_clear_cmd_line_args):
    opts = AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args().get("integrations", {})
    assert opts.get("http_proxy") == "http://example.com:3000"
    assert opts.get("https_proxy") == "https://example.com:3000"
    assert opts.get("timeout") == "50"


def test_global_integrations_options_commented_out(fx_clear_cmd_line_args):
    opts = AppArgumentParser(config_file=mock_paths.MOCK_COMMENTED_APP_CONFIG).parse_args().get("integrations", {})
    assert opts.get("http_proxy") is None
    assert opts.get("https_proxy") is None
    assert opts.get("timeout") is None


def test_log_max_bytes_settings(fx_clear_cmd_line_args):

    # Test reading from app.config
    opts = AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()
    assert isinstance(opts.get("log_max_bytes"), int)
    assert opts.get("log_max_bytes") == 20000000

    # Test default value - commented out in app.config
    opts = AppArgumentParser(config_file=mock_paths.MOCK_COMMENTED_APP_CONFIG).parse_args()
    assert isinstance(opts.get("log_max_bytes"), int)
    assert opts.get("log_max_bytes") == AppArgumentParser.DEFAULT_LOG_MAX_BYTES # default

    # Test overwriting
    sys.argv.extend(["--log_max_bytes", "30000000"])
    opts = AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()
    assert isinstance(opts.get("log_max_bytes"), int)
    assert opts.get("log_max_bytes") == 30000000

    # Test if over limit
    sys.argv.extend(["--log_max_bytes", str(MIN_LOG_BYTES-1)])
    with pytest.raises(ValueError, match=r"log_max_bytes must be either 0 or >= .*"):
        AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()


def test_log_backup_count_settings(fx_clear_cmd_line_args):

    # Test reading from app.config
    opts = AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()
    assert isinstance(opts.get("log_backup_count"), int)
    assert opts.get("log_backup_count") == 11

    # Test default value - commented out in app.config
    opts = AppArgumentParser(config_file=mock_paths.MOCK_COMMENTED_APP_CONFIG).parse_args()
    assert isinstance(opts.get("log_backup_count"), int)
    assert opts.get("log_backup_count") == AppArgumentParser.DEFAULT_LOG_BACKUP_COUNT # default

    # Test overwriting
    sys.argv.extend(["--log_backup_count", "5"])
    opts = AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()
    assert isinstance(opts.get("log_backup_count"), int)
    assert opts.get("log_backup_count") == 5

    # Test if over limit
    sys.argv.extend(["--log_backup_count", str(MIN_BACKUP_COUNT-1)])
    with pytest.raises(ValueError, match="log_backup_count must be a positive value"):
        AppArgumentParser(config_file=mock_paths.MOCK_APP_CONFIG).parse_args()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2020. All Rights Reserved.

from .constants import (APP_CONFIG_LOG_BACKUP_COUNT, APP_CONFIG_LOG_MAX_BYTES,
                        MAX_NUM_WORKERS, MIN_BACKUP_COUNT, MIN_LOG_BYTES,
                        MIN_NUM_WORKERS)

"""Contains a dict to validate the app configs"""

VALIDATE_DICT = {
    "num_workers": {
        "required": False,
        "valid_condition": lambda c: True if c >= MIN_NUM_WORKERS and c <= MAX_NUM_WORKERS else False,
        "invalid_msg": "num_workers must be in the range {} <= {}".format(MIN_NUM_WORKERS, MAX_NUM_WORKERS)
    },
    APP_CONFIG_LOG_MAX_BYTES: {
        "required": False,
        "valid_condition": lambda c: True if c == 0 or c >= MIN_LOG_BYTES else False,
        "invalid_msg": "{} must be either 0 or >= {}".format(APP_CONFIG_LOG_MAX_BYTES, MIN_LOG_BYTES)
    },
    APP_CONFIG_LOG_BACKUP_COUNT: {
        "required": False,
        "valid_condition": lambda c: True if c >= MIN_BACKUP_COUNT else False,
        "invalid_msg": "{} must be a positive value".format(APP_CONFIG_LOG_BACKUP_COUNT)
    }
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2020. All Rights Reserved.

import copy

from .shared_mock_data import mock_constants
from resilient import constants as res_constants
from resilient_circuits import constants, rest_helper


def test_resilient_client_has_custom_headers(fx_mock_resilient_client):
    headers = fx_mock_resilient_client.headers

    assert headers.get("content-type") == "application/json"
    assert headers.get(res_constants.HEADER_MODULE_VER_KEY) == res_constants.HEADER_MODULE_VER_VALUE
    assert headers.get(constants.HEADER_CIRCUITS_VER_KEY) == constants.HEADER_CIRCUITS_VER_VALUE


def test_resilient_client_make_headers_has_custom_header(fx_mock_resilient_client):
    headers = fx_mock_resilient_client.make_headers()
    assert headers.get("content-type") == "application/json"
    assert headers.get(res_constants.HEADER_MODULE_VER_KEY) == res_constants.HEADER_MODULE_VER_VALUE
    assert headers.get(constants.HEADER_CIRCUITS_VER_KEY) == constants.HEADER_CIRCUITS_VER_VALUE


def test_resilient_client_retry_args():

    if rest_helper.resilient_client:
        rest_helper.resilient_client = None

    opts = copy.deepcopy(mock_constants.MOCK_OPTS)

    opts.update({
        res_constants.APP_CONFIG_MAX_CONNECTION_RETRIES: 10,
        res_constants.APP_CONFIG_REQUEST_MAX_RETRIES: 10,
        res_constants.APP_CONFIG_REQUEST_RETRY_DELAY: 10,
        res_constants.APP_CONFIG_REQUEST_RETRY_BACKOFF: 10
    })

    res_client = rest_helper.get_resilient_client(opts)

    assert res_client.max_connection_retries == 10
    assert res_client.request_max_retries == 10
    assert res_client.request_retry_delay == 10
    assert res_client.request_retry_backoff == 10

def test_get_resilient_server_version(fx_mock_resilient_client):

    const = fx_mock_resilient_client.get_const()
    assert "major" in const.get("server_version", {})
    assert "minor" in const.get("server_version", {})



# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

# This here for backwards compatibility
# Some customize.py files have:
# from resilient_circuits.util import *
from resilient import (Definition,
                       TypeDefinition,
                       MessageDestinationDefinition,
                       FunctionDefinition,
                       ActionDefinition,
                       ImportDefinition)

from .resilient_config import get_config_data
from .resilient_customize import get_customization_definitions, get_function_definition



# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

import pkg_resources
try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    __version__ = None

from .actions_component import ResilientComponent
from .action_message import ActionMessageBase, ActionMessage, \
    FunctionMessage, FunctionResult, FunctionError, \
    StatusMessage, BaseFunctionError
from .decorators import function, inbound_app, app_function, handler, required_field, required_action_field, defer, debounce
from .actions_test_component import SubmitTestAction, SubmitTestFunction, SubmitTestInboundApp
from .app_function_component import AppFunctionComponent
from .helpers import is_this_a_selftest



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Circuits component for Action Module subscription and message handling"""

import logging
import threading
from collections import namedtuple
import inspect as _inspect
from functools import wraps
from types import GeneratorType
from circuits import Timer, task, Event
import circuits.core.handlers
from resilient_lib import ResultPayload, validate_fields
from resilient_circuits import constants
from resilient_circuits.action_message import FunctionResult, \
    StatusMessage, StatusMessageEvent, \
    FunctionError_, FunctionErrorEvent

LOG = logging.getLogger(__name__)

# for convenience we alias the circuits 'handler'
handler = circuits.core.handlers.handler


class function(object):
    """Creates a Function Handler.

    This decorator can be applied to methods of classes derived from :class:`ResilientComponent`.
    It marks the method as a handler for the events passed as arguments to the :func:`function` decorator.
    Specify the function's API name as parameter to the decorator.
    The function handler will automatically be subscribed to the function's message destination.
    """
    # This is an extended version of circuits.core.handlers:handler

    def __init__(self, *args, **kwargs):
        if len(args) != 1:
            raise ValueError("Usage: @function(api_name)")
        self.names = args
        self.kwargs = kwargs

    def __call__(self, func):
        """Called at decoration time, with the bare function being decorated"""
        LOG.debug("@function %s", func)

        func.handler = True
        func.function = True

        # Circuits properties
        func.names = self.names
        func.priority = self.kwargs.get("priority", 0)
        func.channel = self.kwargs.get("channel", ",".join(["functions.{}".format(name) for name in self.names]))
        func.override = self.kwargs.get("override", False)
        
        # If getfullargspec if available to us 
        if hasattr(_inspect, 'getfullargspec'):
            args = _inspect.getfullargspec(func)[0]
        else:  # fall back to deprecated getargspec
            args = _inspect.getargspec(func)[0]

        if args and args[0] == "self":
            del args[0]
        func.event = getattr(func, "event", bool(args and args[0] == "event"))

        @wraps(func)
        def decorated(itself, event, *args, **kwargs):
            """the decorated function"""
            LOG.debug("decorated")
            function_parameters = event.message.get("inputs", {})

            def _the_task(event, *args, **kwargs):
                return func(itself, event, *args, **kwargs)

            def _call_the_task(evt, **kwds):
                # On the worker thread, call the function, and handle a single or generator result.
                LOG.debug("%s: _call_the_task", threading.currentThread().name)
                result_list = []
                task_result_or_gen = _the_task(evt, *args, **kwds)
                if not isinstance(task_result_or_gen, GeneratorType):
                    task_result_or_gen = [task_result_or_gen]
                for val in task_result_or_gen:
                    if isinstance(val, StatusMessage):
                        # Fire the wrapped status message event to notify resilient
                        LOG.info("[%s] StatusMessage: %s", evt.name, val)
                        itself.fire(StatusMessageEvent(parent=evt, message=val.text))
                    elif isinstance(val, FunctionResult):
                        # Collect the result for return
                        LOG.debug("[%s] FunctionResult: %s", evt.name, val)
                        val.name = evt.name
                        result_list.append(val)
                    elif isinstance(val, Event):
                        # Some other event, just fire it
                        LOG.debug(val)
                        itself.fire(val)
                    elif isinstance(val, FunctionError_):
                        LOG.error("[%s] FunctionError: %s", evt.name, val)
                        itself.fire(FunctionErrorEvent(parent=evt, message=str(val)))
                        evt.success = False
                        return  # Don't wait for more results!
                    elif isinstance(val, Exception):
                        raise val
                    else:
                        # Whatever this is, add it to the results
                        LOG.debug(val)
                        result_list.append(val)
                return result_list

            the_task = task(_call_the_task, event, **function_parameters)
            ret = yield itself.call(the_task, "functionworker")
            xxx = ret.value
            # Return value is the result_list that was yielded from the wrapped function
            yield xxx
        return decorated


class inbound_app(object):
    def __init__(self, *args, **kwargs):
        if len(args) != 1:
            raise ValueError("Usage: @inbound_app(<{0}>)".format(constants.INBOUND_MSG_APP_CONFIG_Q_NAME))
        self.names = args
        self.kwargs = kwargs

    def __call__(self, ia):
        """
        Called at decoration time, with the bare method being decorated

        :param ia: The method to decorate
        :type ia: resilient_circuits.ResilientComponent
        """
        ia.handler = True
        ia.inbound_handler = True

        # Circuits properties
        ia.names = self.names
        ia.priority = self.kwargs.get("priority", 0)
        ia.channel = "{0}.{1}".format(constants.INBOUND_MSG_DEST_PREFIX, self.names[0])
        ia.override = self.kwargs.get("override", False)
        ia.event = True

        @wraps(ia)
        def inbound_app_decorator(itself, event, *args, **kwargs):
            """
            The decorated method

            :param itself: The method to decorate
            :type itself: resilient_circuits.ResilientComponent
            :param event: The Event with the StompFrame and the Message read off the Message Destination
            :type event: resilient_circuits.action_message.InboundMessage
            """

            def _invoke_inbound_app(evt, **kwds):
                """
                The code to call when a method with the decorator `@inbound_app(<inbound_destination_api_name>)`
                is invoked.

                Returns result_list when method with the decorator `@inbound_app(<inbound_destination_api_name>)` is
                finished processing.

                A method that has this handler should yield a str when done
                    -  E.g:
                        `yield "Processing Complete!"`

                :param evt: The Event with the StompFrame and the Message read off the Message Destination
                :type ia: resilient_circuits.action_message.FunctionMessage
                """
                result_list = []
                LOG.debug("Running _invoke_inbound_app in Thread: %s", threading.currentThread().name)

                # Get the required attribute from the message
                message = evt.message
                inbound_action = message.get("action", "Unknown")

                # Invoke the actual Function
                ia_results = ia(itself, evt.message, inbound_action)

                for r in ia_results:
                    LOG.debug(r)
                    result_list.append(r)

                return result_list

            invoke_inbound_app = task(_invoke_inbound_app, event)
            ia_result = yield itself.call(invoke_inbound_app, "functionworker")
            yield ia_result.value

        return inbound_app_decorator


class app_function(object):
    """
    Creates a new :class:`@app_function <app_function>` decorator.

    This decorator can be applied to methods of the :class:`~resilient_circuits.app_function_component.AppFunctionComponent` class.

    It marks the method as a handler for the events passed as arguments to the decorator.

    Specify the function's API name as parameter to the decorator. **It only accepts 1** ``api_name`` **as an argument.**

    The function handler will automatically be subscribed to the function's ``message destination``.
    """

    def __init__(self, *args, **kwargs):
        if len(args) != 1:
            raise ValueError("Usage: @app_function(api_name)")
        self.names = args
        self.kwargs = kwargs

    def __call__(self, fn):
        """
        Called at decoration time, with the bare function being decorated

        :param fn: The function to decorate
        :type fn: resilient_circuits.ResilientComponent
        """
        fn.handler = True
        fn.function = True

        # Circuits properties
        fn.names = self.names
        fn.priority = self.kwargs.get("priority", 0)
        fn.channel = "functions.{0}".format(self.names[0])
        fn.override = self.kwargs.get("override", False)
        fn.event = True

        @wraps(fn)
        def app_function_decorator(itself, event, *args, **kwargs):
            """
            The decorated function

            :param itself: The function to decorate
            :type itself: resilient_circuits.ResilientComponent
            :param event: The Event with the StompFrame and the Message read off the Message Destination
            :type event: resilient_circuits.action_message.FunctionMessage
            """
            function_inputs = event.message.get("inputs", {})

            def _invoke_app_function(evt, **kwds):
                """
                The code to call when a function with the decorator `@app_function(api_name)`
                is invoked.

                Returns result_list when function with the decorator `@app_function(api_name)` is
                finished processing.

                A method that has this handler should yield a StatusMessage or a FunctionResult
                    -   When a StatusMessage is yield'ed a StatusMessageEvent is fired with the text of the StatusMessage
                    -   When a FunctionResult is yield'ed it calls resilient-lib.ResultPayload.done() with the parameters of
                        FunctionResult being passed to it and appends the result to result_list. E.g:
                            `yield FunctionResult({"key":"value"})`
                            `yield FunctionResult({"key": "value"}, success=False, reason="Bad call")`

                :param evt: The Event with the StompFrame and the Message read off the Message Destination
                :type fn: resilient_circuits.action_message.FunctionMessage
                """
                LOG.debug("Running _invoke_app_function in Thread: %s", threading.currentThread().name)

                result_list = []

                # Validate the fn_inputs in the Message
                fn_inputs = validate_fields([], kwds)
                LOG.info("[%s] Validated function inputs", evt.name)
                LOG.debug("[%s] fn_inputs: %s", evt.name, fn_inputs)

                rp = ResultPayload(itself.PACKAGE_NAME, version=constants.APP_FUNCTION_PAYLOAD_VERSION, **fn_inputs)

                fn_inputs_tuple = namedtuple("fn_inputs", fn_inputs.keys())(*fn_inputs.values())

                # Set evt.message in local thread storage
                itself.set_fn_msg(evt.message)

                # Invoke the actual Function
                fn_results = fn(itself, fn_inputs_tuple)

                for r in fn_results:
                    if isinstance(r, StatusMessage):
                        LOG.info("[%s] StatusMessage: %s", evt.name, r)
                        itself.fire(StatusMessageEvent(parent=evt, message=r.text))

                    elif isinstance(r, FunctionResult):
                        r.name = evt.name
                        if not r.custom_results:
                            r.value = rp.done(
                                content=r.value,
                                success=r.success,
                                reason=r.reason)
                        LOG.info("[%s] Returning results", r.name)
                        result_list.append(r)

                    elif isinstance(r, Exception):
                        raise r

                    else:
                        # Whatever this is, add it to the results
                        LOG.debug(r)
                        result_list.append(r)

                return result_list

            invoke_app_function = task(_invoke_app_function, event, **function_inputs)
            fn_result = yield itself.call(invoke_app_function, "functionworker")
            yield fn_result.value

        return app_function_decorator


class required_field(object):
    """Decorator, declares a required field for a ResilientComponent or its methods"""
    def __init__(self, fieldname, input_type=None):
        self.fieldname = fieldname
        self.input_type = input_type

    def __call__(self, func):
        """Called at decoration time"""
        # Set or extend the function's "custom_fields" attribute
        func.required_fields = getattr(func, "required_fields", {})
        func.required_fields[self.fieldname] = self.input_type
        # The decorated function is unchanged
        return func


class required_action_field(object):
    """Decorator, declares a required field for a ResilientComponent or its methods"""
    def __init__(self, fieldname, input_type=None):
        self.fieldname = fieldname
        self.input_type = input_type

    def __call__(self, func):
        """Called at decoration time"""
        # Set or extend the function's "action_fields" attribute
        func.required_action_fields = getattr(func,
                                              "required_action_fields", {})
        func.required_action_fields[self.fieldname] = self.input_type
        # The decorated function is unchanged
        return func


class defer(object):
    """Decorator for an event handler, delays it awhile.

       Usage:
       Decorate a Resilient Circuits handler.
       This decorator should go *before* the `@handler(...)`.
       Do not use on 'generic' handlers, only on named-event handlers.

       .. code-block:: python

            @defer(delay=5)
            @handler("actions_event_name")
            def _function(self, event, *args, **kwargs):
                # handle the event
                pass
    """
    def __init__(self, *args, **kwargs):
        self.delay = kwargs.get("delay", None)
        if len(args) > 0:
            raise Exception("Usage: @defer() or @defer(delay=<seconds>)")

    def __call__(self, func):
        """Called at decoration time, with function"""
        LOG.debug("@defer %s", func)

        @wraps(func)
        def decorated(itself, event, *args, **kwargs):
            """the decorated function"""
            LOG.debug("decorated")
            if event.defer(itself, delay=self.delay):
                # OK, let's handle it later
                return
            return func(itself, event, *args, **kwargs)
        return decorated


def debounce_get_incident_key(event):
    """Callback to return the debounce-key for an event.
       Multiple events with this key will be debounced together.
       Default is: event name and incident id.
    """
    key = "{} for {}".format(event.name, event.message["incident"]["id"])
    return key


class debounce(object):
    """Decorator for an event handler, debounces multiple occurrences.

       :param delay: (seconds).  Time before the event will be processed.
                  If another event (with the same key) occurs in this
                  time, the timer starts over.
       :param discard: (Boolean, optional).  If true, when there are multiple
                  events before the delay expires, only the most recent one
                  is processed, and the previous ones are discarded.

       Usage:
       Decorate a Resilient Circuits handler.
       This decorator should go *before* the `@handler(...)`.
       Do not use on 'generic' handlers, only on named-event handlers.

       .. code-block:: python

            @debounce(delay=10, discard=True)
            @handler("actions_event_name")
            def _my_action_handler(self, event, *args, **kwargs):
                # handle the event
                pass

    """
    def __init__(self, *args, **kwargs):
        self.delay = kwargs.get("delay", 1)
        self.discard = kwargs.get("discard", False)
        self.get_key = kwargs.get("get_key_func", debounce_get_incident_key)
        self.debouncedata = {}
        if len(args) > 0:
            raise Exception("Usage: @debounce(delay=<seconds>, [discard=True])")

    def __call__(self, func):
        """Called at decoration time, with function"""
        LOG.debug("@debounce %s", func)

        @wraps(func)
        def decorated(itself, event, *args, **kwargs):
            """the decorated function"""
            LOG.debug("decorated")
            # De-bounce messages for this event and the same key:
            # (key is the incident-id, by default):
            # - Don't handle the message immediately.
            #   - Note that we have a deferred event.
            #   - Defer it for <<delay>>.
            # - If an event arrives and there is any deferred message,
            #   - Reset the timer interval to <<delay>>
            #   - Optionally: throw away the new message.
            #     Otherwise: defer this one too (to be processed
            #     immediately after the first deferred message).
            key = self.get_key(event)
            if event.deferred:
                # We deferred this event earlier,
                # and now it has fired without being reset in the meantime.
                # All the pending events are OK to go!  Forget their timers!
                LOG.info("Handling deferred %s", key)
                event.deferred = False
                self.debouncedata.pop(key, None)
            else:
                # This is a new event.
                # Are there any other deferred events for this [action+incident]?
                if key not in self.debouncedata:
                    # We'll keep a list of all the timers
                    self.debouncedata[key] = []
                else:
                    # Duplicate event
                    if self.discard:
                        # Unregister all the previous timers so they don't fire
                        for timer in self.debouncedata[key]:
                            evt = timer.event
                            LOG.debug("Unregister timer")
                            timer.unregister()
                            if evt:
                                # The timer's event will not fire now.
                                # Mark it as not 'deferred' and fire a 'success' child event
                                # so that it gets ack'd to the message queue.
                                LOG.debug("Fire success")
                                evt.deferred = False
                                channels = getattr(evt, "success_channels", evt.channels)
                                itself.fire(evt.child("success", evt, evt.value.value), *channels)
                        # Now we can get rid of the list of timers
                        self.debouncedata[key] = []
                    else:
                        # Reset all the pending timers
                        for timer in self.debouncedata[key]:
                            timer.reset(interval=self.delay)
                # Defer this new event with a timer.
                LOG.info("Deferring %s", key)
                timer = Timer(self.delay, event)
                timer.register(itself)
                event.deferred = True
                # Remember the new timer so that we can reset it if necessary
                self.debouncedata[key].append(timer)
                # We're done until the timer fires
                return
            return func(itself, event, *args, **kwargs)
        return decorated



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import pytest
from mock import patch
from argparse import Namespace
from pkg_resources import EggInfoDistribution, EntryPoint
from resilient_circuits import helpers
from resilient_circuits.cmds import selftest
from tests.shared_mock_data import mock_paths


MOCKED_SELFTEST_ENTRYPOINTS = [
    EntryPoint.parse('test = tests.selftest_tests.mocked_fail_script:selftest', dist=EggInfoDistribution(
        project_name='test',
        version="0.1",
    )), 
    EntryPoint.parse('test2 = tests.selftest_tests.mocked_success_script:selftest', dist=EggInfoDistribution(
        project_name='test2',
        version="0.1"
    )),
    EntryPoint.parse('test3 = tests.selftest_tests.mocked_unimplemented_script:selftest', dist=EggInfoDistribution(
        project_name='test3',
        version="0.1"
    ))
]


def test_error_connecting_to_soar_rest(caplog):
    with pytest.raises(SystemExit) as sys_exit:
        selftest.error_connecting_to_soar("mock_host", status_code=20)

    assert sys_exit.type == SystemExit
    assert sys_exit.value.code == 20


def test_check_soar_rest_connection():
    # TODO
    pass


def test_check_soar_stomp_connection():
    # TODO
    pass


def test_run_apps_selftest_success():
    """
    Test a full run through of selftest with only passing selftest functions
    After calling selftest no exceptions should be raised
    """
    app_configs = helpers.get_configs(path_config_file=mock_paths.MOCK_APP_CONFIG, ALLOW_UNRECOGNIZED=True)
    parser = Namespace(install_list=['test2'], project_name=['test2'])

    with patch("resilient_circuits.bin.resilient_circuits_cmd.pkg_resources.iter_entry_points", create=True) as mocked_entrypoints:
        mocked_entrypoints.return_value = [MOCKED_SELFTEST_ENTRYPOINTS[1]]
        with patch("resilient_circuits.bin.resilient_circuits_cmd.get_config_file") as mocked_config:
            mocked_config.return_value = mock_paths.MOCK_APP_CONFIG
            selftest.run_apps_selftest(parser, app_configs)


def test_run_apps_selftest_failure():
    """
    Test that when selftest is called and a selftest function fails or raises an exception
    that the program exits and gives an error code of 1
    """

    app_configs = helpers.get_configs(path_config_file=mock_paths.MOCK_APP_CONFIG, ALLOW_UNRECOGNIZED=True)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        parser = Namespace(install_list=['test'], project_name=['test'])

        with patch("resilient_circuits.cmds.selftest.pkg_resources.iter_entry_points", create=True) as mocked_entrypoints:
            mocked_entrypoints.return_value = MOCKED_SELFTEST_ENTRYPOINTS
            selftest.run_apps_selftest(parser, app_configs)

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

def test_run_apps_selftest_unimplemented():
    """
    Test that when selftest is called and the app's selftest function returns 
    unimplemented, that the circuits cmd exists and gives an error code of 2
    """

    app_configs = helpers.get_configs(path_config_file=mock_paths.MOCK_APP_CONFIG, ALLOW_UNRECOGNIZED=True)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        parser = Namespace(install_list=['test3'], project_name=['test3'])

        with patch("resilient_circuits.cmds.selftest.pkg_resources.iter_entry_points", create=True) as mocked_entrypoints:
            mocked_entrypoints.return_value = [MOCKED_SELFTEST_ENTRYPOINTS[2]]
            selftest.run_apps_selftest(parser, app_configs)

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 2


# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

""" Circuits component for handling Stomp Connection """

import logging
import ssl
import time
import sys
import traceback
from circuits import BaseComponent, Timer
from circuits.core.handlers import handler
from stompest.config import StompConfig
from stompest.protocol import StompSpec, StompSession
from stompest.sync import Stomp
from stompest.error import StompConnectionError, StompError, StompProtocolError
from stompest.sync.client import LOG_CATEGORY
from resilient_circuits.stomp_events import *
from resilient_circuits.stomp_transport import EnhancedStompFrameTransport
from resilient_circuits import constants


StompSpec.DEFAULT_VERSION = '1.2'
ACK_CLIENT_INDIVIDUAL = StompSpec.ACK_CLIENT_INDIVIDUAL
ACK_AUTO = StompSpec.ACK_AUTO
ACK_CLIENT = StompSpec.ACK_CLIENT
ACK_MODES = (ACK_CLIENT_INDIVIDUAL, ACK_AUTO, ACK_CLIENT)

DEFAULT_MAX_RECONNECT_ATTEMPTS = 3
DEFAULT_STARTUP_MAX_RECONNECT_ATTEMPTS = 3

LOG = logging.getLogger(__name__)


class StompClient(BaseComponent):

    channel = "stomp"

    def init(self, host, port, username=None, password=None,
             connect_timeout=3, connected_timeout=3,
             version=StompSpec.VERSION_1_2, accept_versions=["1.0", "1.1", "1.2"],
             heartbeats=(0, 0), ssl_context=None,
             use_ssl=True,
             key_file=None,
             cert_file=None,
             ca_certs=None,
             ssl_version=ssl.PROTOCOL_TLS if sys.version_info.major == 2 else ssl.PROTOCOL_TLS_CLIENT,
             key_file_password=None,
             proxy_host=None,
             proxy_port=None,
             proxy_user=None,
             proxy_password=None,
             channel=channel,
             stomp_params=None,
             stomp_max_connection_errors=constants.STOMP_MAX_CONNECTION_ERRORS):
        """ Initialize StompClient.  Called after __init__ """
        self.channel = channel
        if proxy_host:
            LOG.info("Connect to %s:%s through proxy %s:%d", host, port, proxy_host, proxy_port)
        else:
            LOG.info("Connect to %s:%s", host, port)

        if use_ssl and not ssl_context:

            ssl_params = dict(key_file=key_file,
                              cert_file=cert_file,
                              ca_certs=ca_certs,
                              ssl_version=ssl_version,
                              password=key_file_password)
            LOG.info("Request to use old-style socket wrapper: %s", ssl_params)
            ssl_context = ssl_params

        if use_ssl:
            uri = "ssl://%s:%s" % (host, port)
        else:
            uri = "tcp://%s:%s" % (host, port)

        # Configure failover options so it only tries based on settings
        # build any parameters passed
        # every connection has at least these two: maxReconnectAttempts, startupMaxReconnectAttempts
        items = [item.split("=", 2) for item in stomp_params.split(",")] if stomp_params else None
        connection_params = {item[0].strip():item[1].strip() for item in items} if items else {}

        if "maxReconnectAttempts" not in connection_params:
            connection_params['maxReconnectAttempts'] = DEFAULT_MAX_RECONNECT_ATTEMPTS
        if "startupMaxReconnectAttempts" not in connection_params:
            connection_params['startupMaxReconnectAttempts'] = DEFAULT_STARTUP_MAX_RECONNECT_ATTEMPTS

        self._stomp_server = "failover:({0})?{1}".format(uri, ",".join(["{}={}".format(k, v) for k, v in connection_params.items()]))
        LOG.debug("Stomp uri: {}".format(self._stomp_server))

        self._stomp_config = StompConfig(uri=self._stomp_server, sslContext=ssl_context,
                                         version=version,
                                         login=username,
                                         passcode=password)

        self._heartbeats = heartbeats
        self._accept_versions = accept_versions
        self._connect_timeout = connect_timeout
        self._connected_timeout = connected_timeout
        Stomp._transportFactory = EnhancedStompFrameTransport
        Stomp._transportFactory.proxy_host = proxy_host
        Stomp._transportFactory.proxy_port = proxy_port
        Stomp._transportFactory.proxy_user = proxy_user
        Stomp._transportFactory.proxy_password = proxy_password
        self._client = Stomp(self._stomp_config)
        self._subscribed = {}
        self.server_heartbeat = None
        self.client_heartbeat = None
        self.last_heartbeat = 0
        self.ALLOWANCE = 2  # multiplier for heartbeat timeouts
        self._stomp_max_connection_errors =  stomp_max_connection_errors # count the number of consecutive errors
        self._stomp_connection_errors = 0

    @property
    def connected(self):
        if self._client.session:
            return self._client.session.state == StompSession.CONNECTED
        else:
            return False

    @property
    def socket_connected(self):
        try:
            if self._client._transport:
                return True
        except:
            pass
        return False

    @property
    def subscribed(self):
        return self._subscribed.keys()

    @property
    def stomp_logger(self):
        return LOG_CATEGORY

    @handler("Disconnect")
    def _disconnect(self, receipt=None, flush=True, reconnect=False):
        try:
            if flush:
                self._subscribed = {}
            if self.connected:
                self._client.disconnect(receipt=receipt)
        except Exception as e:
            LOG.error("Failed to disconnect client")
        try:
            self.fire(Disconnected(reconnect=reconnect))
            self._client.close(flush=flush)
        except Exception as e:
            LOG.error("Failed to close client connection")

        return "disconnected"

    def start_heartbeats(self):
        LOG.info("Client HB: %s  Server HB: %s", self._client.clientHeartBeat, self._client.serverHeartBeat)
        if self._client.clientHeartBeat:
            if self.client_heartbeat:
                # Timer already exists, just reset it
                self.client_heartbeat.reset()
            else:
                LOG.info("Client will send heartbeats to server")
                # Send heartbeats at 80% of agreed rate
                self.client_heartbeat = Timer((self._client.clientHeartBeat / 1000.0) * 0.8,
                                              ClientHeartbeat(), persist=True)
                self.client_heartbeat.register(self)
        else:
            LOG.info("No Client heartbeats will be sent")

        if self._client.serverHeartBeat:
            if self.server_heartbeat:
                # Timer already exists, just reset it
                self.server_heartbeat.reset()
            else:
                LOG.info("Requested heartbeats from server.")
                # Allow a grace period on server heartbeats
                self.server_heartbeat = Timer((self._client.serverHeartBeat / 1000.0) * self.ALLOWANCE,
                                              ServerHeartbeat(), persist=True)
                self.server_heartbeat.register(self)
        else:
            LOG.info("Expecting no heartbeats from Server")

    @handler("Connect")
    def connect(self, event, host=None, *args, **kwargs):
        """ connect to Stomp server """
        LOG.info("Connect to Stomp...")
        try:
            self._client.connect(heartBeats=self._heartbeats,
                                 host=host,
                                 versions=self._accept_versions,
                                 connectTimeout=self._connect_timeout,
                                 connectedTimeout=self._connected_timeout)
            LOG.debug("State after Connection Attempt: %s", self._client.session.state)
            if self.connected:
                LOG.info("Connected to %s", self._stomp_server)
                self.fire(Connected())
                self.start_heartbeats()
                self._stomp_connection_errors = 0 # restart counter
                return "success"

        except StompConnectionError as err:
            LOG.debug(traceback.format_exc())
            # is this error unrecoverable?
            if "no more data" in str(err).lower():
                self._stomp_connection_errors += 1
                if self._stomp_max_connection_errors and self._stomp_connection_errors >= self._stomp_max_connection_errors:
                    LOG.error("Exiting due to unrecoverable error")
                    sys.exit(1) # this will exit resilient-circuits

            self.fire(ConnectionFailed(self._stomp_server))
            event.success = False

        # This logic is added to trap the situation where resilient-circuits does not reconnect from a loss of connection
        #   with the resilient server. In these cases, this error is not survivable and it's best to kill resilient-circuits.
        #   If resilient-circuits is running as a service, it will restart and state would clear for a new stomp connection.
        except StompProtocolError as err:
            LOG.error(traceback.format_exc())
            LOG.error("Exiting due to unrecoverable error")
            sys.exit(1) # this will exit resilient-circuits
        return "fail"

    @handler("ServerHeartbeat")
    def check_server_heartbeat(self, event):
        """ Confirm that heartbeat from server hasn't timed out """
        now = time.time()
        self.last_heartbeat = self._client.lastReceived or self.last_heartbeat
        if self.last_heartbeat:
            elapsed = now-self.last_heartbeat
        else:
            elapsed = -1
        if ((self._client.serverHeartBeat / 1000.0) * self.ALLOWANCE + self.last_heartbeat) < now:
            LOG.error("Server heartbeat timeout. %d seconds since last heartbeat.", elapsed)
            event.success = False
            self.fire(HeartbeatTimeout(now))

    @handler("ClientHeartbeat")
    def send_heartbeat(self, event):
        if self.connected:
            LOG.debug("Sending heartbeat")
            try:
                self._client.beat()
            except (StompConnectionError, StompError) as err:
                event.success = False
                self.fire(OnStompError(None, err))

    @handler("generate_events")
    def generate_events(self, event):
        event.reduce_time_left(0.1)
        if not self.connected:
            return
        try:
            if self._client.canRead(0):
                frame = self._client.receiveFrame()
                LOG.debug("Recieved frame %s", frame)
                if frame.command == StompSpec.ERROR:
                    self.fire(OnStompError(frame, None))
                else:
                    self.fire(Message(frame))
        except (StompConnectionError, StompError) as err:
            LOG.error("Failed attempt to generate events.")
            self.fire(OnStompError(None, err))

    @handler("Send")
    def send(self, event, destination, body, headers=None, receipt=None):
        LOG.debug("send()")
        try:
            self._client.send(destination, body=body.encode('utf-8'), headers=headers, receipt=receipt)
            LOG.debug("Message sent")
        except (StompConnectionError, StompError) as err:
            LOG.error("Error sending frame")
            event.success = False
            self.fire(OnStompError(None, err))
            raise  # To fire Send_failure event

    @handler("Subscribe")
    def _subscribe(self, event, destination, additional_headers=None, ack=ACK_CLIENT_INDIVIDUAL):
        if ack not in ACK_MODES:
            raise ValueError("Invalid client ack mode specified")
        if destination in self._client.session._subscriptions:
            LOG.debug("Ignoring subscribe request to %s. Already subscribed.", destination)
        LOG.info("Subscribe to message destination %s", destination)
        try:
            headers = {StompSpec.ACK_HEADER: ack,
                       'id': destination}
            if additional_headers:
                headers.update(additional_headers)

            # Set ID to match destination name for easy reference later
            frame, token = self._client.subscribe(destination,
                                                  headers)
            self._subscribed[destination] = token
        except (StompConnectionError, StompError) as err:
            LOG.error("Failed to subscribe to queue.")
            event.success = False
            LOG.debug(traceback.format_exc())
            self.fire(OnStompError(None, err))
        # This logic is added to trap the situation where resilient-circuits does not reconnect from a loss of connection
        #   with the resilient server. In these cases, this error is not survivable and it's best to kill resilient-circuits.
        #   If resilient-circuits is running as a service, it will restart and state would clear for a new stomp connection.
        except StompProtocolError as err:
            LOG.error(traceback.format_exc())
            LOG.error("Exiting due to unrecoverable error")
            sys.exit(1) # this will exit resilient-circuits

    @handler("Unsubscribe")
    def _unsubscribe(self, event, destination):
        if destination not in self._subscribed:
            LOG.error("Unsubscribe Request Ignored. Not subscribed to %s", destination)
            return
        try:
            token = self._subscribed.pop(destination)
            frame = self._client.unsubscribe(token)
            LOG.debug("Unsubscribed: %s", frame)
        except (StompConnectionError, StompError) as err:
            event.success = False
            LOG.error("Unsubscribe Failed.")
            self.fire(OnStompError(frame, err))

    @handler("Message")
    def on_message(self, event, headers, message, queue):
        LOG.debug("Stomp message received")

    @handler("Ack")
    def ack_frame(self, event, frame):
        LOG.debug("ack_frame()")
        try:
            self._client.ack(frame)
            LOG.debug("Ack Sent")
        except (StompConnectionError, StompError) as err:
            LOG.error("Error sending ack")
            event.success = False
            self.fire(OnStompError(frame, err))
            raise  # To fire Ack_failure event

    def get_subscription(self, frame):
        """ Get subscription from frame """
        _, token = self._client.message(frame)
        return self._subscribed[token]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

import logging

from mock import MagicMock, patch
from resilient_circuits.app_restartable import ConfigFileUpdateHandler


def test_reload_config_not_already_loading():
    magic_app = MagicMock()
    config_update_handler = ConfigFileUpdateHandler(magic_app)
    with patch("resilient_circuits.app_restartable.AppArgumentParser.parse_args") as mock_parse_args:
        mock_parse_args.return_value = {"loglevel": "DEBUG"}
        magic_app.reloading = False
        magic_app.config_file = None
        config_update_handler.reload_config()
        assert config_update_handler.app.reloading

def test_reload_config_is_already_loading():
    magic_app = MagicMock()
    config_update_handler = ConfigFileUpdateHandler(magic_app)
    with patch("resilient_circuits.app_restartable.AppArgumentParser.parse_args") as mock_parse_args:
        mock_parse_args.return_value = {"loglevel": "DEBUG"}
        magic_app.reloading = True
        config_update_handler.reload_config()
        assert config_update_handler.app.reloading


def test_reset_loglevel():
    config_update_handler = ConfigFileUpdateHandler(MagicMock())
    assert logging.getLevelName(logging.getLogger().level) == "DEBUG" # NOTE: this always work locally, but when run in Tox, this is the log level here

    opts = {"loglevel": "INFO"}
    config_update_handler.reset_loglevel(opts)
    assert logging.getLevelName(logging.getLogger().level) == "INFO"


# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Utility to import customizations"""

from __future__ import print_function

import base64
import json
import logging
from distutils.util import strtobool

import pkg_resources
import resilient
from resilient import (ActionDefinition, Definition, FunctionDefinition,
                       ImportDefinition, MessageDestinationDefinition,
                       SimpleHTTPException, TypeDefinition)
from resilient import helpers as res_helpers
from resilient_circuits.app import AppArgumentParser

try:
    from builtins import input
except ImportError:
    # Python 2
    from __builtin__ import raw_input as input


LOG = logging.getLogger(__name__)


def get_customization_definitions(package):
    """Read the default configuration-data section from the given package"""
    data = None
    try:
        dist = pkg_resources.get_distribution(package)
        entries = pkg_resources.get_entry_map(dist, "resilient.circuits.customize")
        if entries:
            for entry in iter(entries):
                func = entries[entry].load()
                data = func(client=None)
    except pkg_resources.DistributionNotFound:
        pass
    return data or []


def get_function_definition(package, function_name):
    """Find a function in the default configuration-data section from the given package"""
    for definition in get_customization_definitions(package):
        if isinstance(definition, FunctionDefinition):
            new_functions = definition.value
            if not isinstance(new_functions, (tuple, list)):
                new_functions = [new_functions]
            for func in new_functions:
                if func["name"] == function_name:
                    # Found it!  Success.
                    return func
        elif isinstance(definition, ImportDefinition):
            import_data = json.loads(base64.b64decode(definition.value).decode("utf-8"))
            for func in import_data.get("functions"):
                if func["name"] == function_name:
                    # Found it!  Success.
                    return func




def setdefault(dictionary, defaults):
    """Fill in the blanks"""
    for key in defaults.keys():
        if dictionary.get(key) is None:
            dictionary[key] = defaults[key]


def type_displayname(typename):
    """A readable displayname for a type

        >>> type_displayname('__function')
        'Function'
        >>> type_displayname('my_data_table')
        'MyDataTable'
    """
    return typename.title().replace("_", "")


def customize_resilient(args):
    """import customizations to the resilient server"""
    parser = AppArgumentParser(config_file=resilient.get_config_file())
    (opts, extra) = parser.parse_known_args()
    client = resilient.get_client(opts)

    # Call each of the 'customize' entry points to get type definitions,
    # then apply them to the resilient server
    entry_points = pkg_resources.iter_entry_points('resilient.circuits.customize')
    do_customize_resilient(client, entry_points, args.yflag, args.install_list)


def do_customize_resilient(client, entry_points, yflag, install_list):
    """import customizations to the resilient server"""
    ep_count = 0
    customizations = Customizations(client, yflag)
    for entry in entry_points:
        ep_count = ep_count + 1
        def_count = 0
        dist = entry.dist
        dist_str = entry.dist.project_name

        if install_list is None or dist_str in install_list:
            if install_list is not None:
                install_list.remove(dist_str)
            try:
                func = entry.load()
            except ImportError:
                LOG.exception(u"Customizations for package '%s' cannot be loaded.", repr(dist))
                continue

            # The entrypoint function should be a generator
            # that produces Definitions in the sequence required:
            # usually that means:
            # - fields first,
            # - then message destinations,
            # - then actions, functions, etc

            LOG.info(u"Package '%s':", dist)
            definitions = func(client=client)
            for definition in definitions:
                def_count = def_count + 1
                if not isinstance(definition, Definition):
                    pass
                try:
                    if isinstance(definition, ImportDefinition):
                        customizations.load_import(definition, dist)
                    elif isinstance(definition, TypeDefinition):
                        customizations.load_types(definition)
                    elif isinstance(definition, MessageDestinationDefinition):
                        customizations.load_message_destinations(definition)
                    elif isinstance(definition, ActionDefinition):
                        customizations.load_actions(definition)
                    elif isinstance(definition, FunctionDefinition):
                        customizations.load_functions(definition)
                    else:
                        LOG.error(u"Not implemented: %s", type(definition))
                except SimpleHTTPException:
                    LOG.error(u"Failed, %s", customizations.doing)
                    raise
            LOG.info(u"Package '%s' done.", dist)

    if install_list is not None and len(install_list) > 0:
        LOG.warning("%s not found. Check package name(s)", install_list)

    if ep_count == 0:
        LOG.info(u"No customizations are defined by installed packages.")


class Customizations(object):
    """Main class for managing customizations"""

    def __init__(self, client, yflag):
        super(Customizations, self).__init__()

        self.client = client
        self.prompt = not yflag
        self.doing = ""

    def confirm(self, activity):
        """Prompt, if self.prompt"""
        yes = True
        self.doing = u"importing {}".format(activity)
        if self.prompt:
            yes = False
            inp = input(u"    OK to import {}? (y/n):".format(activity))
            try:
                yes = strtobool(inp)
            except ValueError:
                pass
        if not yes:
            print(u"    Not importing {}".format(activity))
        return yes

    def load_import(self, definition, dist):
        """Load an importable block of customizations"""
        import_data = json.loads(base64.b64decode(definition.value).decode("utf-8"))
        import_data = res_helpers.remove_tag(import_data)
        LOG.debug(json.dumps(import_data, indent=2))
        uri = "/configurations/imports"
        done = False
        if self.confirm(u"customizations from '{}'".format(dist)):
            result = self.client.post(uri, import_data)
            import_id = result["id"]
            LOG.debug(result)
            for message in result.get("messages", []):
                LOG.info(u"    %s: %s", message["type"], message["text"])
            if result["status"] == "PENDING":
                if self.confirm(u""):
                    result["status"] = "ACCEPTED"
                    done = True
                else:
                    result["status"] = "REJECTED"
                uri = "/configurations/imports/{}".format(import_id)
                self.client.put(uri, result)

        if done:
            # For each message destination in the import:
            # Set this API account to be a user of the message destination
            # (otherwise the function won't be able to connect!)

            def update_user(dest):
                # Callback for get/put to update the user list
                if self.client.user_id is None:
                    # We are using API key
                    if self.client.api_key_handle not in dest["api_keys"]:
                        LOG.info(u"    Adding api key to message destination {}".format(dest["programmatic_name"]))
                        dest["api_keys"].append(self.client.api_key_handle)
                else:
                    # We are using user/password to authenticate
                    if self.client.user_id not in dest["users"]:
                        LOG.info(u"    Adding user to message destination {}".format(dest["programmatic_name"]))
                        dest["users"].append(self.client.user_id)
                return dest

            uri = "/message_destinations"
            all_destinations = dict((dest["programmatic_name"], dest) for dest in self.client.get(uri)["entities"])
            [dest.get("programmatic_name") for dest in import_data.get("message_destinations", [])]
            for dest in import_data.get("message_destinations", []):
                dest_name = dest.get("programmatic_name")
                if dest_name in all_destinations:
                    dest_id = all_destinations[dest_name]["id"]
                    uri = "/message_destinations/{}".format(dest_id)
                    self.client.get_put(uri, update_user)

    def load_message_destinations(self, definition):
        """Load one or more message destinations"""
        new_destinations = definition.value
        if not isinstance(new_destinations, (tuple, list)):
            new_destinations = [new_destinations]
        uri = "/message_destinations"
        existing_destinations = self.client.get(uri)["entities"]
        existing_destination_names = [dest["programmatic_name"] for dest in existing_destinations]
        existing_destination_dnames = [dest["name"] for dest in existing_destinations]
        for dest in new_destinations:
            if dest["programmatic_name"] in existing_destination_names:
                LOG.info(u"    Message destination exists: %s", dest["programmatic_name"])
            elif dest["name"] in existing_destination_dnames:
                LOG.info(u"    Message destination exists: %s", dest["name"])
            else:
                setdefault(dest, {
                    "name": dest["programmatic_name"],
                    "expect_ack": True,
                    "destination_type": 0
                })
                # Set this API account to be a user of the message destination
                dest["users"] = [self.client.user_id]
                # Don't re-use id
                if "id" in dest:
                    dest.remove("id")
                # Create the message destination
                if self.confirm(u"message destination '{}'".format(dest["programmatic_name"])):
                    self.client.post(uri, dest)
                    LOG.info(u"    Message destination created: %s ('%s')", dest["programmatic_name"], dest["name"])

    def load_types(self, definition):
        """Load a type definition (action fields, incident fields, data table, etc)"""
        new_types = definition.value
        type_name = new_types["type_name"]
        new_fields = new_types["fields"]
        uri = "/types/{0}".format(type_name)
        try:
            existing_types = self.client.get(uri)
        except SimpleHTTPException:
            # There is no type (data-table) of this name.
            # So let's create it.
            uri = "/types"
            if self.confirm(u"type '{}'".format(new_types["type_name"])):
                self.client.post(uri, new_types)
                LOG.info(u"    Type created: %s ('%s')", new_types["type_name"], new_types["display_name"])
            return
        existing_fields = existing_types["fields"]
        for fielduuid in new_fields:
            field = new_fields[fielduuid]
            fieldname = field["name"]
            if fieldname in existing_fields.keys():
                # Merge the field values and update
                LOG.info(u"    %s field exists: %s", type_displayname(type_name), fieldname)
                new_values = field.get("values", [])
                # Remove any ids from the new stuff
                for value in new_values:
                    if "id" in value:
                        value.remove("id")
                new_ids = [value["label"] for value in new_values]
                old_values = existing_fields[fieldname].get("values", [])
                old_ids = [value["label"] for value in old_values]
                # If there are any old values that are not in the new values (by id),
                # add them to the end of the list, and mark as no longer enabled.
                for old_id in set(old_ids) - set(new_ids):
                    old_value = [value for value in old_values if value["label"] == old_id][0]
                    if old_value.get("enabled", True):
                        old_value["enabled"] = False
                        new_values.append(old_value)
                        new_ids.append(old_id)
                # If there are any old values that are in the new values (by label),
                # set their id.
                for value in new_values:
                    label = value["label"]
                    old_values = [value for value in old_values if value["label"] == label]
                    if len(old_values) > 0:
                        value["value"] = old_values[0]["value"]
                    elif "value" in value:
                        value.pop("value", None)
                if new_ids != old_ids:
                    # Post the update
                    LOG.debug(json.dumps(field, indent=2))
                    existing_id = existing_fields[fieldname]["id"]
                    if self.confirm(u"values for {} field '{}'".format(type_displayname(type_name), fieldname)):
                        uri = "/types/{0}/fields/{1}".format(type_name, existing_id)
                        self.client.put(uri, field)
                        LOG.info(u"    %s field updated: %s ('%s')",
                                 type_displayname(type_name), fieldname, field["text"])
            else:
                # Don't re-use id
                if "id" in field:
                    field.remove("id")
                # Create the field
                fields_uri = "/types/{0}/fields".format(type_name)
                if self.confirm(u"{} field '{}'".format(type_displayname(type_name), fieldname)):
                    self.client.post(fields_uri, field)
                    LOG.info(u"    %s field created: %s ('%s')", type_displayname(type_name), fieldname, field["text"])

    def load_actions(self, definition):
        """Load custom actions"""
        new_actions = definition.value
        if not isinstance(new_actions, (tuple, list)):
            new_actions = [new_actions]
        uri = "/actions?handle_format=names"
        existing_actions = self.client.get(uri)["entities"]
        existing_action_names = [action["name"] for action in existing_actions]
        for action in new_actions:
            if action["name"] in existing_action_names:
                LOG.info(u"    Action exists: %s", action["name"])
            else:
                # Don't re-use id
                if "id" in action:
                    action.pop("id", None)
                # Create the action
                if self.confirm(u"action '{}'".format(action["name"])):
                    self.client.post(uri, action)
                    LOG.info(u"    Action created: %s", action["name"])

    def load_functions(self, definition):
        """Load custom functions"""
        new_functions = definition.value
        if not isinstance(new_functions, (tuple, list)):
            new_functions = [new_functions]
        uri = "/functions"
        existing_functions = self.client.get(uri)["entities"]
        existing_function_names = [function["name"] for function in existing_functions]
        for function in new_functions:
            if function["name"] in existing_function_names:
                LOG.info(u"    Function exists: %s", function["name"])
            else:
                setdefault(function, {
                    "display_name": function["name"]
                })
                # Don't re-use id
                if "id" in function:
                    function.pop("id", None)
                # Create the function
                if self.confirm(u"function '{}'".format(function["name"])):
                    self.client.post(uri, function)
                    LOG.info(u"    Function created: %s", function["name"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

import json

from mock import patch
from resilient_circuits.util.resilient_customize import Customizations
from resilient import helpers as res_helpers


def test_customizations_load_import(fx_simple_client, fx_read_mock_definition):
    """
    This is a real basic test just confirming that our move of res_helpers.remove_tag works.
    We test resilient.remove_tag separately
    """

    simple_client = fx_simple_client[0]
    requests_adapter = fx_simple_client[1]
    import_def = fx_read_mock_definition

    simple_client.get_put = lambda x, y: True

    mock_uri = "{0}/rest/orgs/{1}/configurations/imports".format(simple_client.base_url, simple_client.org_id)
    requests_adapter.register_uri('POST', mock_uri, status_code=200, text=json.dumps({"id": 123, "status": "PENDING"}))

    mock_uri = "{0}/rest/orgs/{1}/configurations/imports/123".format(simple_client.base_url, simple_client.org_id)
    requests_adapter.register_uri('PUT', mock_uri, status_code=200, text=json.dumps({"id": 123, "status": "DONE"}))

    mock_uri = "{0}/rest/orgs/{1}/message_destinations".format(simple_client.base_url, simple_client.org_id)
    requests_adapter.register_uri('GET', mock_uri, status_code=200, text=json.dumps({"entities": [{"id": 1, "programmatic_name": "fn_main_mock_integration"}]}))

    c = Customizations(simple_client, False)

    with patch("resilient_circuits.util.resilient_customize.Customizations.confirm") as mock_prompt:
        mock_prompt.return_value = True

        with patch("resilient.helpers.remove_tag", new=res_helpers.remove_tag) as mock_remove_tag:

            c.load_import(import_def, "mock instance")



#!/usr/bin/env python
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.
from __future__ import print_function

"""Resilient Circuits Action Test Tool"""
#
# Usage:
#  $ res-action-test
#

import cmd
import time
import threading
import socket
import select
import sys
import struct
import argparse
import signal
import json
import shlex
import resilient
from resilient_circuits.app import AppArgumentParser
from resilient_circuits.actions_test_component import DEFAULT_TEST_HOST, DEFAULT_TEST_PORT

try:
    from queue import Queue, Empty
except:
    from Queue import Queue, Empty


def get_input(datatype, prompt, value):
    if value is None:
        if sys.version_info[0] >= 3:
            value = input(prompt)
        else:
            value = raw_input(prompt)
            value = value.decode(sys.stdin.encoding)
    else:
        print("{} {}".format(prompt, value))
    if value == "":
        return None
    if datatype == "number":
        value = int(value)
    elif datatype in ["datepicker", "datetimepicker"]:
        value = int(value)
    elif datatype == "boolean":
        value = value.lower()[0:1] in ["y", "t", "1"]
    elif datatype == "textarea":
        value = {"type": "text", "content": value}
    elif datatype == "select":
        value = {"id": 123, "name": value}
    elif datatype == "multiselect":
        value = [{"id": 123, "name": value}]
    return value


class ConnectionThread(threading.Thread):
    """Thread class managing the TCP connection"""

    def __init__(self, host, port, message_queue, response_callback):
        super(ConnectionThread, self).__init__()
        self._stop_event = threading.Event()
        self.message_queue = message_queue
        self.response_callback = response_callback
        self.host = host
        self.port = port

        self.connect()

    def connect(self):
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_address = (self.host, self.port)
        self.client.connect(server_address)

    def reconnect(self):
        self.response_callback(("Connection to Test Server has died.  "
                                "Trying to reconnect..."))
        self.client.close()
        while not self.stopped():
            try:
                self.connect()
                self.response_callback("Connection Restored.")
                return
            except:
                time.sleep(3)

    def run(self):
        while not self.stopped():
            readables, writables, errorables = select.select([self.client],
                                                             [], [self.client],
                                                             1)
            if errorables:
                self.reconnect()
                continue
            if readables:
                msg_len = self.client.recv(4)
                if not msg_len:
                    self.reconnect()
                    continue
                msg_len = struct.unpack('>I', msg_len)[0]
                data = self.client.recv(msg_len).decode()
                self.response_callback(data)
            try:
                msg, msg_id = self.message_queue.get_nowait()
                message = self.format_message(msg, msg_id)
                self.client.send(message)
            except Empty:
                pass

        self.response_callback("Closing connection")
        self.client.close()

    def format_message(self, message, message_id):
        """ Prefix message with length and id and append newline """
        message = self.as_bytes(message + "\n")
        message = struct.pack('>I', message_id) + message
        message = struct.pack('>I', len(message)) + message

        return message

    def as_bytes(self, data):
        return data.encode()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()


class ResilientTestProcessor(cmd.Cmd):
    """Resilient Circuits Test Tool Command Processor"""
    intro = ("Welcome to the Resilient Circuits Action Test Tool. "
             "Type help or ? to list commands.\n")
    prompt = "(restest) "

    def __init__(self, host=DEFAULT_TEST_HOST, port=DEFAULT_TEST_PORT):
        cmd.Cmd.__init__(self)
        self.message_queue = Queue()
        self.response_queue = Queue()
        self.message_id = 0
        try:
            self.conn_thread = ConnectionThread(host, port, self.message_queue,
                                            self.print_response)
        except socket.error:
            print("ERROR: Unable to connect.\n"
                  "The test client requires resilient-circuits already running, with the '--test-actions' option.\n")
            raise
        signal.signal(signal.SIGINT, self.killed)
        signal.signal(signal.SIGTERM, self.killed)
        self.conn_thread.start()

    def postloop(self):
        self.conn_thread.stop()
        self.conn_thread.join()
        print("")

    #
    # -------- Test Commands --------
    #
    def do_submit(self, arg):
        """Submit action from specified queue:  submit <destination_queue> <action_json> """
        queue, action_json = self._parse_arg(arg)
        self._submit_action(queue, action_json)
    # end do_submit

    def do_submitfile(self, arg):
        """Submit action stored in a file from specified queue:
           submitfile <destination_queue> <file containing action_json>
        """
        queue, filename = self._parse_arg(arg)
        try:
            with open(filename) as action_file:
                action_json = action_file.read()
                self._submit_action(queue, action_json)
        except Exception as e:
            print(e)
    # end do_submitfile

    def do_runlist(self, arg):
        """Run a list of commands from a file:  runlist <filename>"""
        if not arg:
            print("runlist command requires a filename")
        try:
            with open(arg) as f:
                self.cmdqueue.extend(f.read().splitlines())
        except Exception as e:
            print("Bad command list file: %s" % str(e))
    # end do_runlist

    def do_quit(self, *args):
        """Quit the Test Tool and exit:  quit"""
        print(("You are exiting the Resilient Circuits "
               "Action Test Tool.  Goodbye.\n"""))
        self.conn_thread.stop()
        return True
    # end do_quit

    def do_EOF(self, *args):
        """ Ctrl+D """
        return self.do_quit()

    def do_function(self, arg):
        """Execute a function"""
        if not arg:
            print("function command requires a function-name")
            return

        parser = AppArgumentParser(config_file=resilient.get_config_file())
        (opts, more) = parser.parse_known_args()
        client = resilient.get_client(opts)

        args = iter(shlex.split(arg))
        try:
            function_name = next(args)
            function_def = client.get("/functions/{}?handle_format=names".format(function_name))
            param_defs = dict({fld["uuid"]: fld for fld in client.get("/types/__function/fields?handle_format=names")})
            function_params = {}
            for param in function_def["view_items"]:
                param_uuid = param["content"]
                param_def = param_defs[param_uuid]
                prompt = "{} ({}, {}): ".format(param_def["name"], param_def["input_type"], param_def["tooltip"])
                try:
                    arg = next(args)
                except StopIteration:
                    arg = None
                function_params[param_def["name"]] = get_input(param_def["input_type"], prompt, arg)

            action_message = {
                "function": {
                    "name": function_name
                },
                "inputs": function_params
            }
            message = json.dumps(action_message, indent=2)
            print(message)
            self._submit_action("function", message)
        except Exception as e:
            print(e)
    # end do_function

    def killed(self, *args):
        """ Handle kill signal """
        self.do_quit()
        self.conn_thread.join()
        raise SystemExit(0)

    # -------- Helping Methods --------
    def print_response(self, data):
        """ print a Test Server response """
        print("\n" + data)
        print(self.prompt, end="")
        sys.stdout.flush()

    def _submit_action(self, queue, action_json):
        """ Send command out over tcp connection """
        if not action_json:
            print("Action required\n")
            return
        self.message_id = self.message_id + 1
        self.message_queue.put((queue + ' ' + action_json, self.message_id))
    # end submit_action

    def _parse_arg(self, arg):
        """ Parse command arguments """
        two_parts = arg.split(' ', 1)
        if len(two_parts) != 2:
            print("Invalid command. Type help or ? to list commands.\n")
            return None, None

        return two_parts
    # end parse_arg


def main():
    description = 'Resilient Circuits Action Test Tool'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--host', dest='host', default=DEFAULT_TEST_HOST,
                        help=("hostname where resilient_circuits program "
                              "is running. Defaults to localhost."))
    parser.add_argument('--port', dest='port', type=int, default=DEFAULT_TEST_PORT,
                        help=("port where resilient_circuits action "
                              "test server is listening. defaults to 8008"))
    args = parser.parse_args()
    ResilientTestProcessor(host=args.host, port=args.port).cmdloop()


if __name__ == '__main__':
    main()
